# manager_context.py
import asyncio
from typing import List, Tuple


from pysnmp.entity.rfc3413 import ntfrcv
from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity import engine, config
from pysnmp.proto.api import v2c

# Set the port to 1622 instead of 162, because 162 is a
# privileged port and requires root access
MANAGER_PORT = 1622


async def start_manager(
    message_count: List[int],
) -> Tuple[SnmpEngine, ntfrcv.NotificationReceiver]:
    # Create SNMP engine
    snmpEngine = engine.SnmpEngine()

    # Transport setup

    # UDP over IPv4
    config.add_transport(
        snmpEngine,
        udp.DOMAIN_NAME,
        udp.UdpTransport().open_server_mode(("localhost", MANAGER_PORT)),
    )

    # SNMPv1/2c setup

    # SecurityName <-> CommunityName mapping
    config.add_v1_system(snmpEngine, "public", "public")

    # SNMPv3/USM setup

    # user: usr-md5-des, auth: MD5, priv DES
    config.add_v3_user(
        snmpEngine,
        "usr-md5-des",
        config.USM_AUTH_HMAC96_MD5,
        "authkey1",
        config.USM_PRIV_CBC56_DES,
        "privkey1",
    )

    # user: usr-md5-des, auth: MD5, priv DES, securityEngineId: 8000000001020304
    # this USM entry is used for TRAP receiving purposes
    config.add_v3_user(
        snmpEngine,
        "usr-md5-des",
        config.USM_AUTH_HMAC96_MD5,
        "authkey1",
        config.USM_PRIV_CBC56_DES,
        "privkey1",
        securityEngineId=v2c.OctetString(hexValue="8000000001020304"),
    )

    # user: usr-none-none, auth: NONE, priv: NONE
    # this USM entry is used for TRAP receiving purposes
    config.add_v3_user(
        snmpEngine,
        "usr-none-none",
        config.USM_AUTH_NONE,
        config.USM_PRIV_NONE,
        securityEngineId=v2c.OctetString(hexValue="8000000001020305"),
    )

    # Callback function for receiving notifications
    # noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
    def cbFun(
        snmpEngine, stateReference, contextEngineId, contextName, varBinds, cbCtx
    ):
        message_count[0] += 1

    # Register SNMP Application at the SNMP engine
    receiver = ntfrcv.NotificationReceiver(snmpEngine, cbFun)

    # Run I/O dispatcher which would receive queries and send confirmations
    snmpEngine.transport_dispatcher.job_started(1)  # this job would never finish

    snmpEngine.open_dispatcher()

    # Wait for the manager to start
    await asyncio.sleep(1)

    # return the engine
    return snmpEngine, receiver


class ManagerContextManager:
    """
    A context manager for managing the lifecycle of an SNMP test manager.

    Usage:
    async with ManagerContextManager() as manager:
        # Perform operations with the manager

    When the context is entered, the manager is started using the `start_manager()` function.
    When the context is exited, the manger's transport dispatcher is stopped and closed.

    Note: The `start_manager()` function and the `transportDispatcher` attribute are not defined in this code snippet.
    """

    manager: SnmpEngine
    receiver: ntfrcv.NotificationReceiver
    message_count: List[int]

    async def __aenter__(self):
        self.message_count = [0]
        self.manager, self.receiver = await start_manager(self.message_count)
        return self.manager, self.message_count

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.receiver.close(self.manager)

        self.manager.transport_dispatcher.job_finished(1)
        self.manager.close_dispatcher()
        self.manager._close()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import random
from hashlib import md5, sha1

from pyasn1.type import univ
from pysnmp.proto import errind, error
from pysnmp.proto.secmod.rfc3414 import localkey
from pysnmp.proto.secmod.rfc3414.auth import hmacmd5, hmacsha
from pysnmp.proto.secmod.rfc3414.priv import base
from pysnmp.proto.secmod.rfc7860.auth import hmacsha2

PysnmpCryptoError = False
try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.decrepit.ciphers import algorithms
    from cryptography.hazmat.primitives.ciphers import Cipher, modes
except ImportError:
    PysnmpCryptoError = True

random.seed()


# 8.2.4


class Des(base.AbstractEncryptionService):
    """DES for USM (RFC3414)."""

    SERVICE_ID = (1, 3, 6, 1, 6, 3, 10, 1, 2, 2)  # usmDESPrivProtocol
    KEY_SIZE = 16

    local_int = random.randrange(0, 0xFFFFFFFF)

    def hash_passphrase(self, authProtocol, privKey) -> univ.OctetString:
        """Hash a passphrase.

        This method hashes a passphrase using the authentication protocol.

        Parameters
        ----------
        authProtocol : tuple
            The authentication protocol OID.
        privKey : bytes
            The privacy key.

        Returns
        -------
        bytes
            The hashed passphrase.
        """
        if authProtocol == hmacmd5.HmacMd5.SERVICE_ID:
            hashAlgo = md5
        elif authProtocol == hmacsha.HmacSha.SERVICE_ID:
            hashAlgo = sha1
        elif authProtocol in hmacsha2.HmacSha2.HASH_ALGORITHM:
            hashAlgo = hmacsha2.HmacSha2.HASH_ALGORITHM[authProtocol]
        else:
            raise error.ProtocolError(f"Unknown auth protocol {authProtocol}")
        return localkey.hash_passphrase(privKey, hashAlgo)

    def localize_key(self, authProtocol, privKey, snmpEngineID) -> bytes:
        """Localize privacy key.

        This method localizes privacy key using the authentication protocol
        and the authoritative SNMP engine ID.

        Parameters
        ----------
        authProtocol : tuple
            The authentication protocol OID.
        privKey : bytes
            The privacy key.
        snmpEngineID : bytes
            The authoritative SNMP engine ID.

        Returns
        -------
        bytes
            The localized privacy key.
        """
        if authProtocol == hmacmd5.HmacMd5.SERVICE_ID:
            hashAlgo = md5
        elif authProtocol == hmacsha.HmacSha.SERVICE_ID:
            hashAlgo = sha1
        elif authProtocol in hmacsha2.HmacSha2.HASH_ALGORITHM:
            hashAlgo = hmacsha2.HmacSha2.HASH_ALGORITHM[authProtocol]
        else:
            raise error.ProtocolError(f"Unknown auth protocol {authProtocol}")
        localPrivKey = localkey.localize_key(privKey, snmpEngineID, hashAlgo)
        return localPrivKey[: self.KEY_SIZE]

    # 8.1.1.1
    def __get_encryption_key(self, privKey, snmpEngineBoots):
        desKey = privKey[:8]
        preIV = privKey[8:16]

        securityEngineBoots = int(snmpEngineBoots)

        salt = [
            securityEngineBoots >> 24 & 0xFF,
            securityEngineBoots >> 16 & 0xFF,
            securityEngineBoots >> 8 & 0xFF,
            securityEngineBoots & 0xFF,
            self.local_int >> 24 & 0xFF,
            self.local_int >> 16 & 0xFF,
            self.local_int >> 8 & 0xFF,
            self.local_int & 0xFF,
        ]
        if self.local_int == 0xFFFFFFFF:
            self.local_int = 0
        else:
            self.local_int += 1

        return (
            desKey.asOctets(),
            univ.OctetString(salt).asOctets(),
            univ.OctetString(
                map(lambda x, y: x ^ y, salt, preIV.asNumbers())
            ).asOctets(),
        )

    @staticmethod
    def __get_decryption_key(privKey, salt):
        return (
            privKey[:8].asOctets(),
            univ.OctetString(
                map(lambda x, y: x ^ y, salt.asNumbers(), privKey[8:16].asNumbers())
            ).asOctets(),
        )

    # 8.2.4.1
    def encrypt_data(self, encryptKey, privParameters, dataToEncrypt):
        """Encrypt data."""
        if PysnmpCryptoError:
            raise error.StatusInformation(errorIndication=errind.encryptionError)

        snmpEngineBoots, snmpEngineTime, salt = privParameters

        # 8.3.1.1
        desKey, salt, iv = self.__get_encryption_key(encryptKey, snmpEngineBoots)

        # 8.3.1.2
        privParameters = univ.OctetString(salt)

        # 8.1.1.2
        plaintext = (
            dataToEncrypt
            + univ.OctetString((0,) * (8 - len(dataToEncrypt) % 8)).asOctets()
        )

        try:
            des = _cryptography_cipher(desKey, iv).encryptor()
            ciphertext = des.update(plaintext) + des.finalize()

        except AttributeError:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedPrivProtocol
            )

        # 8.3.1.3 & 4
        return univ.OctetString(ciphertext), privParameters

    # 8.2.4.2
    def decrypt_data(self, decryptKey, privParameters, encryptedData):
        """Decrypt data."""
        if PysnmpCryptoError:
            raise error.StatusInformation(errorIndication=errind.decryptionError)

        snmpEngineBoots, snmpEngineTime, salt = privParameters

        # 8.3.2.1
        if len(salt) != 8:
            raise error.StatusInformation(errorIndication=errind.decryptionError)

        # 8.3.2.2 no-op

        # 8.3.2.3
        desKey, iv = self.__get_decryption_key(decryptKey, salt)

        # 8.3.2.4 -> 8.1.1.3
        if len(encryptedData) % 8 != 0:
            raise error.StatusInformation(errorIndication=errind.decryptionError)

        try:
            # 8.3.2.6
            des = _cryptography_cipher(desKey, iv).decryptor()
            return des.update(encryptedData.asOctets()) + des.finalize()

        except AttributeError:
            raise error.StatusInformation(
                errorIndication=errind.unsupportedPrivProtocol
            )


def _cryptography_cipher(key, iv):
    """Build a cryptography DES(-like) Cipher object.

    .. note::

        pyca/cryptography does not support DES directly because it is a
        seriously old, insecure, and deprecated algorithm. However,
        triple DES is just three rounds of DES (encrypt, decrypt, encrypt)
        done by taking a key three times the size of a DES key and breaking
        it into three pieces. So triple DES with des_key * 3 is equivalent
        to DES.

    :param bytes key: Encryption key
    :param bytes iv: Initialization vector
    :returns: TripleDES Cipher instance providing DES behavior by using
        provided DES key
    :rtype: cryptography.hazmat.primitives.ciphers.Cipher
    """
    return Cipher(
        algorithm=algorithms.TripleDES(key * 3),
        mode=modes.CBC(iv),
        backend=default_backend(),
    )



"""
TRAP over multiple transports
+++++++++++++++++++++++++++++

The following script sends two SNMP TRAP notification using the
following options:

* with SNMPv1
* with community name 'public'
* over IPv4/UDP and IPv6/UDP
* send TRAP notification
* to a Manager at demo.pysnmp.com:162 and [::1]
* with TRAP ID 'coldStart' specified as an OID
* include managed objects information:
* with default Uptime value
* with default Agent Address with '127.0.0.1'
* overriding Enterprise OID with 1.3.6.1.4.1.20408.4.1.1.2

The following Net-SNMP commands will produce similar SNMP notification:

| $ snmptrap -v1 -c public udp:demo.pysnmp.com 1.3.6.1.4.1.20408.4.1.1.2 127.0.0.1 1 0 12345
| $ snmptrap -v1 -c public udp6:[::1] 1.3.6.1.4.1.20408.4.1.1.2 127.0.0.1 1 0 12345

"""  #
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.carrier.asyncio.dgram import udp, udp6
from pyasn1.codec.ber import encoder
from pysnmp.proto import api

# Protocol version to use
pMod = api.PROTOCOL_MODULES[api.SNMP_VERSION_1]
# pMod = api.PROTOCOL_MODULES[api.SNMP_VERSION_2C]

# Build PDU
trapPDU = pMod.TrapPDU()
pMod.apiTrapPDU.set_defaults(trapPDU)

# Traps have quite different semantics across proto versions
if pMod == api.PROTOCOL_MODULES[api.SNMP_VERSION_1]:
    pMod.apiTrapPDU.set_enterprise(trapPDU, (1, 3, 6, 1, 1, 2, 3, 4, 1))
    pMod.apiTrapPDU.set_generic_trap(trapPDU, "coldStart")

# Build message
trapMsg = pMod.Message()
pMod.apiMessage.set_defaults(trapMsg)
pMod.apiMessage.set_community(trapMsg, "public")
pMod.apiMessage.set_pdu(trapMsg, trapPDU)

transportDispatcher = AsyncioDispatcher()

# UDP/IPv4
transportDispatcher.register_transport(
    udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)
transportDispatcher.send_message(
    encoder.encode(trapMsg), udp.DOMAIN_NAME, ("demo.pysnmp.com", 162)
)

# UDP/IPv6
transportDispatcher.register_transport(
    udp6.DOMAIN_NAME, udp6.Udp6AsyncioTransport().open_client_mode()
)
transportDispatcher.send_message(
    encoder.encode(trapMsg), udp6.DOMAIN_NAME, ("::1", 162)
)

# Dispatcher will finish as all scheduled messages are sent
transportDispatcher.run_dispatcher(3)

transportDispatcher.close_dispatcher()



"""
Multiple concurrent notifications
+++++++++++++++++++++++++++++++++

Send multiple SNMP notifications at once using the following options:

* SNMPv2c and SNMPv3
* with community name 'public'
* over IPv4/UDP
* send TRAP notification
* to multiple Managers
* with TRAP ID 'coldStart' specified as a MIB symbol
* include managed object information specified as var-bind objects pair

Functionally similar to:

| $ snmptrap -v1 -c public demo.pysnmp.com 1.3.6.1.4.1.20408.4.1.1.2 demo.pysnmp.com 6 432 12345 1.3.6.1.2.1.1.1.0 s "my system"
| $ snmptrap -v2c -c public demo.pysnmp.com 123 1.3.6.1.6.3.1.1.5.1

"""  #
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    # List of targets in the following format:
    # ( ( authData, transportTarget ), ... )
    TARGETS = (
        # 1-st target (SNMPv1 over IPv4/UDP)
        (
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create(("demo.pysnmp.com", 162)),
            ContextData(),
        ),
        # 2-nd target (SNMPv2c over IPv4/UDP)
        (
            CommunityData("public"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 162)),
            ContextData(),
        ),
    )

    snmpEngine = SnmpEngine()

    for authData, transportTarget, contextData in TARGETS:
        await send_notification(
            snmpEngine,
            authData,
            transportTarget,
            contextData,
            "trap",  # NotifyType
            NotificationType(ObjectIdentity("SNMPv2-MIB", "coldStart")).add_varbinds(
                ("1.3.6.1.2.1.1.1.0", "my name")
            ),
        )

    snmpEngine.transport_dispatcher.run_dispatcher()


asyncio.run(run())



"""
SNMPv1
++++++

Send SNMP GET request using the following options:

  * with SNMPv1, community 'public'
  * over IPv4/UDP
  * to an Agent at demo.snmplabs.com:161
  * for an instance of SNMPv2-MIB::sysDescr.0 MIB object
  * Based on asyncio I/O framework

Functionally similar to:

| $ snmpget -v1 -c public demo.pysnmp.com SNMPv2-MIB::sysDescr.0

"""  #
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    snmpEngine = SnmpEngine()

    iterator = get_cmd(
        snmpEngine,
        CommunityData("public", mpModel=0),
        await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
        ContextData(),
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
    )

    errorIndication, errorStatus, errorIndex, varBinds = await iterator

    if errorIndication:
        print(errorIndication)

    elif errorStatus:
        print(
            "{} at {}".format(
                errorStatus.prettyPrint(),
                errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
            )
        )
    else:
        for varBind in varBinds:
            print(" = ".join([x.prettyPrint() for x in varBind]))

    snmpEngine.close_dispatcher()


asyncio.run(run())



"""
Specific SNMP Engine ID
+++++++++++++++++++++++

Listen and respond to SNMP GET/SET/GETNEXT/GETBULK queries with
the following options:

* SNMPv3
* with SNMP EngineID: 8000000004030201
* with USM user 'usr-md5-des', auth: MD5, priv DES
* allow access to SNMPv2-MIB objects (1.3.6.1.2.1)
* over IPv4/UDP, listening at 127.0.0.1:161

The following Net-SNMP command will walk this Agent:

| $ snmpwalk -v3 -u usr-md5-des -l authPriv -A authkey1 -X privkey1 -e 8000000004030201 localhost .1.3.6

"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.proto import rfc1902

# Create SNMP engine
snmpEngine = engine.SnmpEngine(rfc1902.OctetString(hexValue="8000000004030201"))

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# SNMPv3/USM setup

# user: usr-md5-des, auth: MD5, priv DES
config.add_v3_user(
    snmpEngine,
    "usr-md5-des",
    config.USM_AUTH_HMAC96_MD5,
    "authkey1",
    config.USM_PRIV_CBC56_DES,
    "privkey1",
)

# Allow full MIB access for each user at VACM
config.add_vacm_user(
    snmpEngine, 3, "usr-md5-des", "authPriv", (1, 3, 6, 1, 2, 1), (1, 3, 6, 1, 2, 1)
)

# Get default SNMP context this SNMP engine serves
snmpContext = context.SnmpContext(snmpEngine)

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()
except:
    snmpEngine.close_dispatcher()
    raise



import asyncio
from datetime import datetime
import pytest
from pysnmp.hlapi.v1arch.asyncio import *
from pysnmp.proto.errind import RequestTimedOut
from pysnmp.proto.rfc1905 import errorStatus as pysnmp_errorStatus
from pysnmp.smi import builder, compiler, view

from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v1_get():
    async with AgentContextManager():
        with SnmpDispatcher() as snmpDispatcher:
            iterator = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ("1.3.6.1.2.1.1.1.0", None),
            )

            errorIndication, errorStatus, errorIndex, varBinds = iterator

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "1.3.6.1.2.1.1.1.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")

            name = pysnmp_errorStatus.namedValues.getName(errorStatus)
            assert name == "noError"


@pytest.mark.asyncio
async def test_v1_get_ipv6():
    async with AgentContextManager(enable_ipv6=True):
        with SnmpDispatcher() as snmpDispatcher:
            iterator = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await Udp6TransportTarget.create(("localhost", AGENT_PORT)),
                ("1.3.6.1.2.1.1.1.0", None),
            )

            errorIndication, errorStatus, errorIndex, varBinds = iterator

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "1.3.6.1.2.1.1.1.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")

            name = pysnmp_errorStatus.namedValues.getName(errorStatus)
            assert name == "noError"


@pytest.mark.asyncio
async def test_v1_get_timeout_invalid_target():
    with SnmpDispatcher() as snmpDispatcher:

        async def run_get():
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("community_string"),
                await UdpTransportTarget.create(("1.2.3.4", 161), timeout=1, retries=0),
                ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.1.0")),
            )
            assert isinstance(errorIndication, RequestTimedOut)
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 0

        start = datetime.now()
        try:
            await asyncio.wait_for(run_get(), timeout=3)
            end = datetime.now()
            elapsed_time = (end - start).total_seconds()
            assert elapsed_time >= 1 and elapsed_time <= 3
        except asyncio.TimeoutError:
            assert False, "Test case timed out"


@pytest.mark.asyncio
async def test_v1_get_timeout_slow_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:

            async def run_get():
                errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                    snmpDispatcher,
                    CommunityData("public", mpModel=0),
                    await UdpTransportTarget.create(
                        ("localhost", AGENT_PORT), timeout=1, retries=0
                    ),
                    ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.1.0")),
                )
                for varBind in varBinds:
                    print([str(varBind[0]), varBind[1]])

            start = datetime.now()
            try:
                await asyncio.wait_for(run_get(), timeout=3)
                end = datetime.now()
                elapsed_time = (end - start).total_seconds()
                assert elapsed_time >= 1 and elapsed_time <= 3
            except asyncio.TimeoutError:
                assert False, "Test case timed out"


@pytest.mark.asyncio
async def test_v1_get_no_access_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.3")),
            )
            assert errorIndication is None
            assert (
                errorStatus.prettyPrint() == "noSuchName"
            )  # v1 does not have noAccess


@pytest.mark.asyncio
async def test_v1_get_dateandtime_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            # Step 1: Set up MIB builder and add custom MIB directory
            mibBuilder = builder.MibBuilder()
            compiler.addMibCompiler(mibBuilder)
            mibViewController = view.MibViewController(mibBuilder)

            # Load the custom MIB
            mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
            snmpDispatcher.cache["mibViewController"] = mibViewController

            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ObjectType(
                    ObjectIdentity("LEXTUDIO-TEST-MIB", "testDateAndTime", 0)
                ),  # "1.3.6.1.4.1.60069.9.5.0"
            )
            assert errorIndication is None
            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert (
                varBinds[0][0].prettyPrint() == "LEXTUDIO-TEST-MIB::testDateAndTime.0"
            )
            assert (
                varBinds[0][1].prettyPrint() == "2024-11-3,19:52:40.0,+0:0"
            )  # IMPORTANT: test DateAndTime display hint.


@pytest.mark.asyncio
async def test_v1_get_physaddress_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            # Step 1: Set up MIB builder and add custom MIB directory
            mibBuilder = builder.MibBuilder()
            compiler.addMibCompiler(mibBuilder)
            mibViewController = view.MibViewController(mibBuilder)

            # Load the custom MIB
            mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
            snmpDispatcher.cache["mibViewController"] = mibViewController

            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ObjectType(
                    ObjectIdentity("LEXTUDIO-TEST-MIB", "testPhysAddress", 0)
                ),  # "1.3.6.1.4.1.60069.9.6.0"
            )
            assert errorIndication is None
            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert (
                varBinds[0][0].prettyPrint() == "LEXTUDIO-TEST-MIB::testPhysAddress.0"
            )
            assert (
                varBinds[0][1].prettyPrint() == "00:11:22:33:44:55"
            )  # IMPORTANT: test PhysAddress display hint.


@pytest.mark.asyncio
async def test_v1_get_scaled_integer_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            # Step 1: Set up MIB builder and add custom MIB directory
            mibBuilder = builder.MibBuilder()
            compiler.addMibCompiler(mibBuilder)
            mibViewController = view.MibViewController(mibBuilder)

            # Load the custom MIB
            mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
            snmpDispatcher.cache["mibViewController"] = mibViewController

            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ObjectType(
                    ObjectIdentity("LEXTUDIO-TEST-MIB", "testScaledInteger", 0)
                ),  # "1.3.6.1.4.1.60069.9.7.0"
            )
            assert errorIndication is None
            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert (
                varBinds[0][0].prettyPrint() == "LEXTUDIO-TEST-MIB::testScaledInteger.0"
            )
            assert (
                varBinds[0][1].prettyPrint() == "5.0"
            )  # IMPORTANT: test display hint "d-1".


@pytest.mark.asyncio
async def test_v1_get_scaled_unsigned_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            # Step 1: Set up MIB builder and add custom MIB directory
            mibBuilder = builder.MibBuilder()
            compiler.addMibCompiler(mibBuilder)
            mibViewController = view.MibViewController(mibBuilder)

            # Load the custom MIB
            mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
            snmpDispatcher.cache["mibViewController"] = mibViewController

            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ObjectType(
                    ObjectIdentity("LEXTUDIO-TEST-MIB", "testScaledUnsigned", 0)
                ),  # "1.3.6.1.4.1.60069.9.8.0"
            )
            assert errorIndication is None
            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert (
                varBinds[0][0].prettyPrint()
                == "LEXTUDIO-TEST-MIB::testScaledUnsigned.0"
            )
            assert (
                varBinds[0][1].prettyPrint() == "0.5"  # GitHub issue #139
            )  # IMPORTANT: test display hint "d-1".


@pytest.mark.asyncio
async def test_v1_get_multiple_objects():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpDispatcher() as snmpDispatcher:
            # # Step 1: Set up MIB builder and add custom MIB directory
            # mibBuilder = builder.MibBuilder()
            # compiler.addMibCompiler(mibBuilder)
            # mibViewController = view.MibViewController(mibBuilder)

            # # Load the custom MIB
            # mibBuilder.loadModules("LEXTUDIO-TEST-MIB")
            # snmpDispatcher.cache["mibViewController"] = mibViewController

            netiftable = [
                ObjectIdentity("IF-MIB", "ifDescr", 1),
                ObjectIdentity("IF-MIB", "ifDescr", 2),
            ]

            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("demo.pysnmp.com", 161), timeout=1, retries=0
                ),
                *[ObjectType(x) for x in netiftable],
            )
            assert errorIndication is None
            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 2
            assert varBinds[0][1].prettyPrint() != varBinds[1][1].prettyPrint()



"""
Implementing conceptual table
+++++++++++++++++++++++++++++

Listen and respond to SNMP GET/SET/GETNEXT/GETBULK queries with
the following options:

* SNMPv2c
* with SNMP community "public"
* define a simple SNMP Table within a newly created EXAMPLE-MIB
* pre-populate SNMP Table with a single row of values
* allow read access only to the subtree where example SNMP Table resides
* over IPv4/UDP, listening at 127.0.0.1:161

The following Net-SNMP commands will populate and walk a table:

| $ snmpset -v2c -c public 127.0.0.1 1.3.6.6.1.5.2.97.98.99 s "my value"
| $ snmpset -v2c -c public 127.0.0.1 1.3.6.6.1.5.4.97.98.99 i 4
| $ snmpwalk -v2c -c public 127.0.0.1 1.3.6

...while the following command will destroy the same row

| $ snmpset -v2c -c public 127.0.0.1 1.3.6.6.1.5.4.97.98.99 i 6
| $ snmpwalk -v2c -c public 127.0.0.1 1.3.6

"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.proto.api import v2c


# Create SNMP engine
snmpEngine = engine.SnmpEngine()

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# SNMPv2c setup

# SecurityName <-> CommunityName mapping.
config.add_v1_system(snmpEngine, "my-area", "public")

# Allow read MIB access for this user / securityModels at VACM
config.add_vacm_user(
    snmpEngine, 2, "my-area", "noAuthNoPriv", (1, 3, 6, 6), (1, 3, 6, 6)
)

# Create an SNMP context
snmpContext = context.SnmpContext(snmpEngine)

# --- define custom SNMP Table within a newly defined EXAMPLE-MIB ---

mibBuilder = snmpContext.get_mib_instrum().get_mib_builder()

(MibTable, MibTableRow, MibTableColumn, MibScalarInstance) = mibBuilder.import_symbols(
    "SNMPv2-SMI", "MibTable", "MibTableRow", "MibTableColumn", "MibScalarInstance"
)

(RowStatus,) = mibBuilder.import_symbols("SNMPv2-TC", "RowStatus")

mibBuilder.export_symbols(
    "__EXAMPLE-MIB",
    # table object
    exampleTable=MibTable((1, 3, 6, 6, 1)).setMaxAccess("read-create"),
    # table row object, also carries references to table indices
    exampleTableEntry=MibTableRow((1, 3, 6, 6, 1, 5))
    .setMaxAccess("read-create")
    .setIndexNames((0, "__EXAMPLE-MIB", "exampleTableColumn1")),
    # table column: string index
    exampleTableColumn1=MibTableColumn(
        (1, 3, 6, 6, 1, 5, 1), v2c.OctetString()
    ).setMaxAccess("read-create"),
    # table column: string value
    exampleTableColumn2=MibTableColumn(
        (1, 3, 6, 6, 1, 5, 2), v2c.OctetString()
    ).setMaxAccess("read-create"),
    # table column: integer value with default
    exampleTableColumn3=MibTableColumn(
        (1, 3, 6, 6, 1, 5, 3), v2c.Integer32(123)
    ).setMaxAccess("read-create"),
    # table column: row status
    exampleTableStatus=MibTableColumn(
        (1, 3, 6, 6, 1, 5, 4), RowStatus("notExists")
    ).setMaxAccess("read-create"),
)

# --- end of custom SNMP table definition, empty table now exists ---

# --- populate custom SNMP table with one row ---

(
    exampleTableEntry,
    exampleTableColumn2,
    exampleTableColumn3,
    exampleTableStatus,
) = mibBuilder.import_symbols(
    "__EXAMPLE-MIB",
    "exampleTableEntry",
    "exampleTableColumn2",
    "exampleTableColumn3",
    "exampleTableStatus",
)
rowInstanceId = exampleTableEntry.getInstIdFromIndices("example record one")
mibInstrumentation = snmpContext.get_mib_instrum()
mibInstrumentation.write_variables(
    (exampleTableColumn2.name + rowInstanceId, "my string value"),
    (exampleTableColumn3.name + rowInstanceId, 123456),
    (exampleTableStatus.name + rowInstanceId, "createAndGo"),
)

# --- end of SNMP table population ---

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()
except:
    snmpEngine.close_dispatcher()
    raise



import pytest
from pysnmp.hlapi.v1arch.asyncio import *
from tests.agent_context import AGENT_PORT, AgentContextManager
from pysnmp.proto.rfc1905 import errorStatus as pysnmp_errorStatus


@pytest.mark.asyncio
async def test_v1_get():
    async with AgentContextManager():
        with Slim(1) as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.get(
                "public",
                "localhost",
                AGENT_PORT,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                retries=0,
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
            assert isinstance(varBinds[0][1], OctetString)

            name = pysnmp_errorStatus.namedValues.getName(errorStatus)
            assert name == "noError"


@pytest.mark.asyncio
async def test_v1_get_old():
    async with AgentContextManager():
        slim = Slim(1)
        errorIndication, errorStatus, errorIndex, varBinds = await slim.get(
            "public",
            "localhost",
            AGENT_PORT,
            ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
        )

        assert errorIndication is None
        assert errorStatus == 0
        assert errorIndex == 0
        assert len(varBinds) == 1
        assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
        assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
        assert isinstance(varBinds[0][1], OctetString)

        slim.close()


@pytest.mark.asyncio
async def test_v1_next():
    async with AgentContextManager():
        with Slim(1) as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.next(
                "public",
                "localhost",  # "demo.pysnmp.com",
                AGENT_PORT,  # 161,
                ObjectType(
                    ObjectIdentity("SNMPv2-MIB", "sysDescr", 0).load_mibs("PYSNMP-MIB")
                ),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"
            assert (
                varBinds[0][1].prettyPrint() == "PYSNMP-MIB::pysnmp"
            )  # IMPORTANT: MIB is needed to resolve this name
            assert type(varBinds[0][1]).__name__ == "ObjectIdentity"


@pytest.mark.asyncio
async def test_v1_set():
    async with AgentContextManager():
        with Slim(1) as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.set(
                "public",
                "localhost",
                AGENT_PORT,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0), "Shanghai"),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysLocation.0"
            assert varBinds[0][1].prettyPrint() == "Shanghai"
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
@pytest.mark.parametrize("num_bulk", [1, 2, 50])
async def test_v2c_bulk(num_bulk):
    async with AgentContextManager():
        with Slim() as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.bulk(
                "public",
                "localhost",
                AGENT_PORT,
                0,
                num_bulk,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                retries=0,
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == num_bulk
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"
            if num_bulk > 1:
                assert varBinds[1][0].prettyPrint() == "SNMPv2-MIB::sysUpTime.0"
            if num_bulk > 2:
                assert varBinds[2][0].prettyPrint() == "SNMPv2-MIB::sysContact.0"


@pytest.mark.asyncio
async def test_v2_get():
    async with AgentContextManager():
        with Slim() as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.get(
                "public",
                "localhost",
                AGENT_PORT,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_v2_next():
    async with AgentContextManager():
        with Slim() as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.next(
                "public",
                "localhost",  # "demo.pysnmp.com",
                AGENT_PORT,  # 161,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"


@pytest.mark.asyncio
async def test_v2_set():
    async with AgentContextManager():
        with Slim() as slim:
            errorIndication, errorStatus, errorIndex, varBinds = await slim.set(
                "public",
                "localhost",
                AGENT_PORT,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0), "Shanghai"),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysLocation.0"
            assert varBinds[0][1].prettyPrint() == "Shanghai"
            assert isinstance(varBinds[0][1], OctetString)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.proto import rfc1905

noSuchObject = rfc1905.noSuchObject  # noqa: N816
noSuchInstance = rfc1905.noSuchInstance  # noqa: N816
endOfMibView = endOfMib = rfc1905.endOfMibView  # noqa: N816



"""
Implementing MIB objects
++++++++++++++++++++++++

This script explains how SNMP Agent application could model
real-world data as Managed Objects defined in MIB.

"""  #
from pysnmp.smi import builder

# MIB Builder is normally pre-created by SNMP engine
mibBuilder = builder.MibBuilder()

#
# This may be done in a stand-alone file and then loaded up
# by SNMP Agent
#

# A base class for a custom Managed Object
(MibScalarInstance,) = mibBuilder.import_symbols("SNMPv2-SMI", "MibScalarInstance")

# Managed object specification
(sysLocation,) = mibBuilder.import_symbols("SNMPv2-MIB", "sysLocation")


# Custom Managed Object
class MySysLocationInstance(MibScalarInstance):
    # noinspection PyUnusedLocal
    def readGet(self, varBind, **context):
        # Just return a custom value
        return varBind[0], self.syntax.clone("The Leaky Cauldron")


sysLocationInstance = MySysLocationInstance(sysLocation.name, (0,), sysLocation.syntax)

# Register Managed Object with a MIB tree
mibBuilder.export_symbols(
    # '__' prefixed MIB modules take precedence on indexing
    "__MY-LOCATION-MIB",
    sysLocationInstance=sysLocationInstance,
)

if __name__ == "__main__":
    #
    # This is what is done internally by Agent.
    #
    from pysnmp.smi import instrum, exval

    mibInstrum = instrum.MibInstrumController(mibBuilder)

    print("Remote manager read access to MIB instrumentation (table walk)")

    varBinds = [((), None)]

    while True:
        varBinds = mibInstrum.read_next_variables(*varBinds)
        oid, val = varBinds[0]
        if exval.endOfMib.isSameTypeWith(val):
            break
        print(oid, val.prettyPrint())



#

# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
"""
This module provides a class for generating sequential integers in a reasonably multi-thread safe manner.

Classes:
    Integer: A class to generate the next integer value within a specified
             range and increment.

Usage example:
    integer_gen = Integer(maximum=1000, increment=256)
    next_value = integer_gen()
"""
import random

random.seed()


class Integer:
    """Return a next value in a reasonably MT-safe manner."""

    def __init__(self, maximum, increment=256):
        """Create a new instance of integer generator."""
        self.__maximum = maximum
        if increment >= maximum:
            increment = maximum
        self.__increment = increment
        self.__threshold = increment // 2
        e = random.randrange(self.__maximum - self.__increment)
        self.__bank = list(range(e, e + self.__increment))

    def __repr__(self):
        """Return a string representation of the object."""
        return "%s(%d, %d)" % (
            self.__class__.__name__,
            self.__maximum,
            self.__increment,
        )

    def __call__(self):
        """Return next integer."""
        v = self.__bank.pop(0)
        if v % self.__threshold:
            return v
        else:
            # this is MT-safe unless too many (~ increment/2) threads
            # bump into this code simultaneously
            e = self.__bank[-1] + 1
            if e > self.__maximum:
                e = 0
            self.__bank.extend(range(e, e + self.__threshold))
            return v



"""
SNMPv1
++++++

Send SNMP GET request using the following options:

  * with SNMPv1, community 'public'
  * over IPv4/UDP
  * to an Agent at demo.pysnmp.com:161
  * for an instance of SNMPv2-MIB::sysDescr.0 MIB object
  * Based on asyncio I/O framework

Functionally similar to:

| $ snmpgetnext -v1 -c public demo.pysnmp.com SNMPv2-MIB::sysDescr.0

"""  #

import pytest
from pysnmp.hlapi.v1arch.asyncio import *
from pysnmp.smi import builder, compiler, view
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v1_next():
    async with AgentContextManager():
        with SnmpDispatcher() as snmpDispatcher:
            errorIndication, errorStatus, errorIndex, varBinds = await next_cmd(
                snmpDispatcher,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ObjectType(
                    ObjectIdentity("SNMPv2-MIB", "sysDescr", 0).load_mibs("PYSNMP-MIB")
                ),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"
            assert (
                varBinds[0][1].prettyPrint() == "PYSNMP-MIB::pysnmp"
            )  # IMPORTANT: MIB is needed to resolve this name
            assert type(varBinds[0][1]).__name__ == "ObjectIdentity"



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.codec.ber import decoder, eoo
from pyasn1.error import PyAsn1Error
from pyasn1.type import univ
from pysnmp.proto.error import ProtocolError


def decode_message_version(wholeMsg):
    """Decode SNMP version from the message."""
    try:
        seq, wholeMsg = decoder.decode(
            wholeMsg,
            asn1Spec=univ.Sequence(),
            recursiveFlag=False,
            substrateFun=lambda a, b, c: (a, b[:c]),
        )
        ver, wholeMsg = decoder.decode(
            wholeMsg,
            asn1Spec=univ.Integer(),
            recursiveFlag=False,
            substrateFun=lambda a, b, c: (a, b[:c]),
        )
        if eoo.endOfOctets.isSameTypeWith(ver):
            raise ProtocolError("EOO at SNMP version component")
        return ver
    except PyAsn1Error as exc:
        raise ProtocolError("Invalid BER at SNMP version component: %s" % exc)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
# License: http://snmplabs.com/pysnmp/license.html
#
"""
This module initializes the asyncio-based high-level API for pysnmp.

It imports all components from the v3arch.asyncio module, which is the
default architecture for the asyncio-based implementation of pysnmp.

Attributes:
    *: All components from the pysnmp.hlapi.v3arch.asyncio module.
"""
from pysnmp.hlapi.v3arch.asyncio import *  # default is v3arch



#!/usr/bin/env python3
"""
Script to publish Python packages for pysmi.
Equivalent to Publish-Packages.ps1
"""

import os
import re
import subprocess
import sys
from pathlib import Path


def main():
    """Main function to publish packages."""
    # Define the path to the .pypirc file
    pypirc_path = Path.home() / ".pypirc"

    # Initialize credential variables
    pypi_username = None
    pypi_password = None

    # Check if the .pypirc file exists
    if not pypirc_path.is_file():
        print(f"Error: ~/.pypirc file not found at '{pypirc_path}'.", file=sys.stderr)
        sys.exit(1)

    print(f"Reading credentials from {pypirc_path}...")

    # Read the file content and parse for [pypi] credentials
    try:
        pypirc_content = pypirc_path.read_text()
        # Simple parsing assuming [pypi] section and username/password lines
        if "[pypi]" in pypirc_content:
            # Parse username
            username_match = re.search(r"(?m)^\s*username\s*=\s*(.+)$", pypirc_content)
            if username_match:
                pypi_username = username_match.group(1).strip()

            # Parse password
            password_match = re.search(r"(?m)^\s*password\s*=\s*(.+)$", pypirc_content)
            if password_match:
                pypi_password = password_match.group(1).strip()
    except Exception as e:
        print(f"Error reading or parsing '{pypirc_path}': {str(e)}", file=sys.stderr)
        sys.exit(1)

    # Validate that credentials were found
    if not pypi_username or not pypi_password:
        print(
            f"Error: Could not find username and/or password under the [pypi] section in '{pypirc_path}'.",
            file=sys.stderr,
        )
        sys.exit(1)

    print(
        f"Credentials found. Username: {pypi_username}"
    )  # Username is usually __token__

    # Construct and execute the uv publish command
    print("Attempting to publish packages using credentials from ~/.pypirc...")
    try:
        subprocess.run(
            ["uv", "publish", "--username", pypi_username, "--password", pypi_password],
            check=True,
        )
        print("Publish command executed.")
    except subprocess.CalledProcessError as e:
        print(f"Error executing uv publish: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Multiple concurrent notifications
+++++++++++++++++++++++++++++++++

Send multiple SNMP notifications at once using the following options:

* SNMPv2c
* with community name 'public'
* over IPv4/UDP
* send INFORM notification
* to multiple Managers
* with TRAP ID 'coldStart' specified as a MIB symbol

Functionally similar to:

| $ snmptrap -v2c -c public demo.pysnmp.com 12345 1.3.6.1.6.3.1.1.5.2
| $ snmpinform -v2c -c public demo.pysnmp.com 12345 1.3.6.1.6.3.1.1.5.2

"""  #

import asyncio
from pysnmp.hlapi.v1arch.asyncio import *


async def sendone(snmpDispatcher, hostname, notifyType):
    iterator = await send_notification(
        snmpDispatcher,
        CommunityData("public"),
        await UdpTransportTarget.create((hostname, 162)),
        notifyType,
        NotificationType(ObjectIdentity("1.3.6.1.6.3.1.1.5.2"))
        .load_mibs("SNMPv2-MIB")
        .add_varbinds(
            ("1.3.6.1.6.3.1.1.4.3.0", "1.3.6.1.4.1.20408.4.1.1.2"),
            ("1.3.6.1.2.1.1.1.0", OctetString("my system")),
        ),
    )

    errorIndication, errorStatus, errorIndex, varBinds = iterator

    if errorIndication:
        print(errorIndication)

    elif errorStatus:
        print(
            "{}: at {}".format(
                errorStatus.prettyPrint(),
                errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
            )
        )

    else:
        for varBind in varBinds:
            print(" = ".join([x.prettyPrint() for x in varBind]))


async def main():
    with SnmpDispatcher() as snmpDispatcher:
        await asyncio.gather(
            sendone(snmpDispatcher, "demo.pysnmp.com", "trap"),
            sendone(snmpDispatcher, "demo.pysnmp.com", "inform"),
        )


asyncio.run(main())



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import warnings


from pysnmp import error
from pysnmp.entity import config

__all__ = [
    "CommunityData",
    "UsmUserData",
    "USM_PRIV_CBC168_3DES",
    "USM_PRIV_CFB128_AES",
    "USM_PRIV_CFB192_AES",
    "USM_PRIV_CFB256_AES",
    "USM_PRIV_CFB192_AES_BLUMENTHAL",
    "USM_PRIV_CFB256_AES_BLUMENTHAL",
    "USM_PRIV_CBC56_DES",
    "USM_AUTH_HMAC96_MD5",
    "USM_AUTH_HMAC96_SHA",
    "USM_AUTH_HMAC128_SHA224",
    "USM_AUTH_HMAC192_SHA256",
    "USM_AUTH_HMAC256_SHA384",
    "USM_AUTH_HMAC384_SHA512",
    "USM_AUTH_NONE",
    "USM_PRIV_NONE",
    "USM_KEY_TYPE_PASSPHRASE",
    "USM_KEY_TYPE_MASTER",
    "USM_KEY_TYPE_LOCALIZED",
]


class CommunityData:
    """Creates SNMP v1/v2c configuration entry.

    This object can be used by
    :py:class:`~pysnmp.hlapi.v3arch.asyncio.AsyncCommandGenerator` or
    :py:class:`~pysnmp.hlapi.v3arch.asyncio.AsyncNotificationOriginator`
    and their derivatives for adding new entries to Local Configuration
    Datastore (LCD) managed by :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
    class instance.

    See :RFC:`2576#section-5.3` for more information on the
    *SNMP-COMMUNITY-MIB::snmpCommunityTable*.

    Parameters
    ----------
    communityIndex: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        Unique index value of a row in snmpCommunityTable. If it is the
        only positional parameter, it is treated as a *communityName*.

    communityName: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        SNMP v1/v2c community string.

    mpModel: :py:class:`int`
        SNMP message processing model AKA SNMP version. Known SNMP versions are:

        * `0` - for SNMP v1
        * `1` - for SNMP v2c (default)


    contextEngineId: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        Indicates the location of the context in which management
        information is accessed when using the community string
        specified by the above communityName.

    contextName: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        The context in which management information is accessed when
        using the above communityName.

    tag: :py:class:`str`
        Arbitrary string that specifies a set of transport endpoints
        from which a command responder application will accept
        management requests with given *communityName* or to which
        notification originator application will send notifications
        when targets are specified by a tag value(s).

        The other way to look at the *tag* feature is that it can make
        specific *communityName* only valid for certain targets.

        The other use-case is when multiple distinct SNMP peers share
        the same *communityName* -- binding each instance of
        *communityName* to transport endpoint lets you distinguish
        SNMP peers from each other (e.g. resolving *communityName* into
        proper *securityName*).

        For more technical information on SNMP configuration tags please
        refer to :RFC:`3413#section-4.1.1` and :RFC:`2576#section-5.3`
        (e.g. the *snmpCommunityTransportTag* object).

        See also: :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget`

    Warnings
    --------
    If the same *communityIndex* value is supplied repeatedly with
    different *communityName* (or other parameters), the later call
    supersedes all previous calls.

    Make sure not to configure duplicate *communityName* values unless
    they have distinct *mpModel* and/or *tag* fields. This will make
    *communityName* based database lookup ambiguous.

    Examples
    --------
    >>> from pysnmp.hlapi.v3arch.asyncio import CommunityData
    >>> CommunityData('public')
    CommunityData(communityIndex='s1410706889', communityName=<COMMUNITY>, mpModel=1, contextEngineId=None, contextName='', tag='')
    >>> CommunityData('public', 'public')
    CommunityData(communityIndex='public', communityName=<COMMUNITY>, mpModel=1, contextEngineId=None, contextName='', tag='')
    >>>

    """

    message_processing_model = 1  # Default is SMIv2
    security_model = message_processing_model + 1
    security_level = "noAuthNoPriv"
    context_name = b""
    tag = b""

    def __init__(
        self,
        communityIndex,
        communityName=None,
        mpModel=None,
        contextEngineId=None,
        contextName=None,
        tag=None,
        securityName=None,
    ):
        """Create a community data object."""
        if mpModel is not None:
            self.message_processing_model = mpModel
            self.security_model = mpModel + 1
        self.contextEngineId = contextEngineId
        if contextName is not None:
            self.context_name = contextName
        if tag is not None:
            self.tag = tag
        # a single arg is considered as a community name
        if communityName is None:
            communityName, communityIndex = communityIndex, None
        self.communityName = communityName
        # Autogenerate communityIndex if not specified
        if communityIndex is None:
            self.communityIndex = self.securityName = "s%s" % hash(
                (
                    self.communityName,
                    self.message_processing_model,
                    self.contextEngineId,
                    self.context_name,
                    self.tag,
                )
            )
        else:
            self.communityIndex = communityIndex
            self.securityName = (
                securityName is not None and securityName or communityIndex
            )

    def __hash__(self):
        """Return a hash value of the object."""
        raise TypeError("%s is not hashable" % self.__class__.__name__)

    def __repr__(self):
        """Return a string representation of the object."""
        return "{}(communityIndex={!r}, communityName=<COMMUNITY>, mpModel={!r}, contextEngineId={!r}, contextName={!r}, tag={!r}, securityName={!r})".format(
            self.__class__.__name__,
            self.communityIndex,
            self.message_processing_model,
            self.contextEngineId,
            self.context_name,
            self.tag,
            self.securityName,
        )

    def clone(
        self,
        communityIndex=None,
        communityName=None,
        mpModel=None,
        contextEngineId=None,
        contextName=None,
        tag=None,
        securityName=None,
    ):
        """Clone the object with new parameters."""
        # a single arg is considered as a community name
        if communityName is None:
            communityName, communityIndex = communityIndex, None
        return self.__class__(
            communityIndex,
            communityName is None and self.communityName or communityName,
            mpModel is None and self.message_processing_model or mpModel,
            contextEngineId is None and self.contextEngineId or contextEngineId,
            contextName is None and self.context_name or contextName,
            tag is None and self.tag or tag,
            securityName is None and self.securityName or securityName,
        )


USM_AUTH_NONE = config.USM_AUTH_NONE
"""No Authentication Protocol"""

USM_AUTH_HMAC96_MD5 = config.USM_AUTH_HMAC96_MD5
"""The HMAC-MD5-96 Digest Authentication Protocol (:RFC:`3414#section-6`)"""

USM_AUTH_HMAC96_SHA = config.USM_AUTH_HMAC96_SHA
"""The HMAC-SHA-96 Digest Authentication Protocol AKA SHA-1 (:RFC:`3414#section-7`)"""

USM_AUTH_HMAC128_SHA224 = config.USM_AUTH_HMAC128_SHA224
"""The HMAC-SHA-2 Digest Authentication Protocols (:RFC:`7860`)"""

USM_AUTH_HMAC192_SHA256 = config.USM_AUTH_HMAC192_SHA256
"""The HMAC-SHA-2 Digest Authentication Protocols (:RFC:`7860`)"""

USM_AUTH_HMAC256_SHA384 = config.USM_AUTH_HMAC256_SHA384
"""The HMAC-SHA-2 Digest Authentication Protocols (:RFC:`7860`)"""

USM_AUTH_HMAC384_SHA512 = config.USM_AUTH_HMAC384_SHA512
"""The HMAC-SHA-2 Digest Authentication Protocols (:RFC:`7860`)"""

USM_PRIV_NONE = config.USM_PRIV_NONE
"""No Privacy Protocol"""

USM_PRIV_CBC56_DES = config.USM_PRIV_CBC56_DES
"""The CBC-DES Symmetric Encryption Protocol (:RFC:`3414#section-8`)"""

USM_PRIV_CBC168_3DES = config.USM_PRIV_CBC168_3DES
"""The 3DES-EDE Symmetric Encryption Protocol (`draft-reeder-snmpv3-usm-3desede-00 <https:://tools.ietf.org/html/draft-reeder-snmpv3-usm-3desede-00#section-5>`_)"""

USM_PRIV_CFB128_AES = config.USM_PRIV_CFB128_AES
"""The CFB128-AES-128 Symmetric Encryption Protocol (:RFC:`3826#section-3`)"""

USM_PRIV_CFB192_AES = config.USM_PRIV_CFB192_AES
"""The CFB128-AES-192 Symmetric Encryption Protocol (`draft-blumenthal-aes-usm-04 <https:://tools.ietf.org/html/draft-blumenthal-aes-usm-04#section-3>`_) with Reeder key localization. Also known as AES-192-Cisco"""

USM_PRIV_CFB256_AES = config.USM_PRIV_CFB256_AES
"""The CFB128-AES-256 Symmetric Encryption Protocol (`draft-blumenthal-aes-usm-04 <https:://tools.ietf.org/html/draft-blumenthal-aes-usm-04#section-3>`_) with Reeder key localization. Also known as AES-256-Cisco"""

USM_PRIV_CFB192_AES_BLUMENTHAL = config.USM_PRIV_CFB192_AES_BLUMENTHAL
"""The CFB128-AES-192 Symmetric Encryption Protocol (`draft-blumenthal-aes-usm-04 <https:://tools.ietf.org/html/draft-blumenthal-aes-usm-04#section-3>`_)"""

USM_PRIV_CFB256_AES_BLUMENTHAL = config.USM_PRIV_CFB256_AES_BLUMENTHAL
"""The CFB128-AES-256 Symmetric Encryption Protocol (`draft-blumenthal-aes-usm-04 <https:://tools.ietf.org/html/draft-blumenthal-aes-usm-04#section-3>`_)"""

USM_KEY_TYPE_PASSPHRASE = config.USM_KEY_TYPE_PASSPHRASE
"""USM key material type - plain-text pass phrase (:RFC:`3414#section-2.6`)"""

USM_KEY_TYPE_MASTER = config.USM_KEY_TYPE_MASTER
"""USM key material type - hashed pass-phrase AKA master key (:RFC:`3414#section-2.6`)"""

USM_KEY_TYPE_LOCALIZED = config.USM_KEY_TYPE_LOCALIZED
"""USM key material type - hashed pass-phrase hashed with Context SNMP Engine ID (:RFC:`3414#section-2.6`)"""


class UsmUserData:
    """Creates SNMP v3 User Security Model (USM) configuration entry.

    This object can be used by
    :py:class:`~pysnmp.hlapi.v3arch.asyncio.AsyncCommandGenerator` or
    :py:class:`~pysnmp.hlapi.v3arch.asyncio.AsyncNotificationOriginator`
    and their derivatives for adding new entries to Local Configuration
    Datastore (LCD) managed by :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
    class instance.

    See :RFC:`3414#section-5` for more information on the
    *SNMP-USER-BASED-SM-MIB::usmUserTable*.

    Parameters
    ----------
    userName: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        A human readable string representing the name of the SNMP USM user.

    Other Parameters
    ----------------
    authKey: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        Initial value of the secret authentication key.  If not set,
        :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_NONE`
        is implied.  If set and no *authProtocol* is specified,
        :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC96_MD5`
        takes effect.

    privKey: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        Initial value of the secret encryption key.  If not set,
        :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_NONE`
        is implied.  If set and no *privProtocol* is specified,
        :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CBC56_DES`
        takes effect.

    authProtocol: :py:class:`tuple`, :py:class:`~pysnmp.proto.rfc1902.ObjectIdentifier`
        An indication of whether messages sent on behalf of this USM user
        can be authenticated, and if so, the type of authentication protocol
        which is used.

        Supported authentication protocol identifiers are:

        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_NONE` (default is *authKey* not given)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC96_MD5` (default if *authKey* is given)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC96_SHA`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC128_SHA224`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC192_SHA256`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC256_SHA384`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_AUTH_HMAC384_SHA512`


    securityEngineId: :py:class:`~pysnmp.proto.rfc1902.OctetString`
        The snmpEngineID of the authoritative SNMP engine to which a
        dateRequest message is to be sent. Will be automatically
        discovered from peer if not given, unless localized keys
        are used. In the latter case *securityEngineId* must be
        specified.

        See :RFC:`3414#section-2.5.1` for technical explanation.

    securityName: :py:class:`str`, :py:class:`~pysnmp.proto.rfc1902.OctetString`
        Together with the snmpEngineID it identifies a row in the
        *SNMP-USER-BASED-SM-MIB::usmUserTable* that is to be used
        for securing the message.

        See :RFC:`3414#section-2.5.1` for technical explanation.

    privProtocol: :py:class:`tuple`, :py:class:`~pysnmp.proto.rfc1902.ObjectIdentifier`
        An indication of whether messages sent on behalf of this USM user
        be encrypted, and if so, the type of encryption protocol which is used.

        Supported encryption protocol identifiers are:

        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_NONE` (default is *privKey* not given)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CBC56_DES` (default if *privKey* is given)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CBC168_3DES`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CFB128_AES`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CFB192_AES`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_PRIV_CFB256_AES`


    authKeyType: :py:class:`int`
        Type of `authKey` material. See :RFC:`3414#section-2.6` for
        technical explanation.

        Supported key types are:

        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_PASSPHRASE` (default)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_MASTER`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_LOCALIZED`

    privKeyType: :py:class:`int`
        Type of `privKey` material. See :RFC:`3414#section-2.6` for
        technical explanation.

        Supported key types are:

        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_PASSPHRASE` (default)
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_MASTER`
        * :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_LOCALIZED`

    Notes
    -----
    If :py:class:`~pysnmp.hlapi.v3arch.asyncio.USM_KEY_TYPE_LOCALIZED` is used when
    running a non-authoritative SNMP engine, USM key localization
    mechanism is not invoked. As a consequence, local SNMP engine
    configuration won't get automatically populated with remote SNMP
    engine's *securityEngineId*.

    Therefore peer SNMP engine's *securityEngineId* must be added
    to local configuration and associated with its localized keys.

    Alternatively, the magic *securityEngineId* value of five zeros
    (*0x0000000000*) can be used to refer to the localized keys that
    should be used with any unknown remote SNMP engine. This feature
    is specific to pysnmp.

    Examples
    --------
    >>> from pysnmp.hlapi.v3arch.asyncio import UsmUserData
    >>> UsmUserData('testuser', authKey='authenticationkey')
    UsmUserData(userName='testuser', authKey=<AUTHKEY>, privKey=<PRIVKEY>, authProtocol=(1,3,6,1,6,3,10,1,1,2), privProtocol=(1,3,6,1,6,3,10,1,2,1))
    >>> UsmUserData('testuser', authKey='authenticationkey', privKey='encryptionkey')
    UsmUserData(userName='testuser', authKey=<AUTHKEY>, privKey=<PRIVKEY>, authProtocol=(1,3,6,1,6,3,10,1,1,2), privProtocol=(1,3,6,1,6,3,10,1,2,2))
    >>>

    """

    authentication_key = privacy_key = None
    authentication_protocol = config.USM_AUTH_NONE
    privacy_protocol = config.USM_PRIV_NONE
    security_level = "noAuthNoPriv"
    security_model = 3
    message_processing_model = 3
    context_name = b""

    def __init__(
        self,
        userName,
        authKey=None,
        privKey=None,
        authProtocol=None,
        privProtocol=None,
        securityEngineId=None,
        securityName=None,
        authKeyType=USM_KEY_TYPE_PASSPHRASE,
        privKeyType=USM_KEY_TYPE_PASSPHRASE,
    ):
        """Create a USM user data object."""
        self.userName = userName
        if securityName is None:
            self.securityName = userName
        else:
            self.securityName = securityName

        if authKey is not None:
            self.authentication_key = authKey
            if authProtocol is None:
                self.authentication_protocol = config.USM_AUTH_HMAC96_MD5
            else:
                self.authentication_protocol = authProtocol
            if self.security_level != "authPriv":
                self.security_level = "authNoPriv"

        if privKey is not None:
            self.privacy_key = privKey
            if self.authentication_protocol == config.USM_AUTH_NONE:
                raise error.PySnmpError("Privacy implies authenticity")
            self.security_level = "authPriv"
            if privProtocol is None:
                self.privacy_protocol = config.USM_PRIV_CBC56_DES
            else:
                self.privacy_protocol = privProtocol

        self.securityEngineId = securityEngineId
        self.authKeyType = authKeyType
        self.privKeyType = privKeyType

    def __hash__(self):
        """Return a hash value of the object."""
        raise TypeError("%s is not hashable" % self.__class__.__name__)

    def __repr__(self):
        """Return a string representation of the object."""
        return "{}(userName={!r}, authKey=<AUTHKEY>, privKey=<PRIVKEY>, authProtocol={!r}, privProtocol={!r}, securityEngineId={!r}, securityName={!r}, authKeyType={!r}, privKeyType={!r})".format(
            self.__class__.__name__,
            self.userName,
            self.authentication_protocol,
            self.privacy_protocol,
            self.securityEngineId is None and "<DEFAULT>" or self.securityEngineId,
            self.securityName,
            self.authKeyType,
            self.privKeyType,
        )

    def clone(
        self,
        userName=None,
        authKey=None,
        privKey=None,
        authProtocol=None,
        privProtocol=None,
        securityEngineId=None,
        securityName=None,
        authKeyType=None,
        privKeyType=None,
    ):
        """Clone the object with new parameters."""
        return self.__class__(
            userName is None and self.userName or userName,
            authKey is None and self.authentication_key or authKey,
            privKey is None and self.privacy_key or privKey,
            authProtocol is None and self.authentication_protocol or authProtocol,
            privProtocol is None and self.privacy_protocol or privProtocol,
            securityEngineId is None and self.securityEngineId or securityEngineId,
            securityName is None and self.securityName or securityName,
            authKeyType is None and self.authKeyType or USM_KEY_TYPE_PASSPHRASE,
            privKeyType is None and self.privKeyType or USM_KEY_TYPE_PASSPHRASE,
        )

    # Compatibility API
    deprecated_attributes = {
        "authKey": "authentication_key",
        "privKey": "privacy_key",
        "authProtocol": "authentication_protocol",
        "privProtocol": "privacy_protocol",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )


# Backward-compatible protocol IDs

usmNoAuthProtocol = USM_AUTH_NONE  # noqa: N816
usmHMACMD5AuthProtocol = USM_AUTH_HMAC96_MD5  # noqa: N816
usmHMACSHAAuthProtocol = USM_AUTH_HMAC96_SHA  # noqa: N816
usmHMAC128SHA224AuthProtocol = USM_AUTH_HMAC128_SHA224  # noqa: N816
usmHMAC192SHA256AuthProtocol = USM_AUTH_HMAC192_SHA256  # noqa: N816
usmHMAC256SHA384AuthProtocol = USM_AUTH_HMAC256_SHA384  # noqa: N816
usmHMAC384SHA512AuthProtocol = USM_AUTH_HMAC384_SHA512  # noqa: N816
usmNoPrivProtocol = USM_PRIV_NONE  # noqa: N816
usmDESPrivProtocol = USM_PRIV_CBC56_DES  # noqa: N816
usm3DESEDEPrivProtocol = USM_PRIV_CBC168_3DES  # noqa: N816
usmAesCfb128Protocol = USM_PRIV_CFB128_AES  # noqa: N816
usmAesCfb192Protocol = USM_PRIV_CFB192_AES  # noqa: N816
usmAesCfb256Protocol = USM_PRIV_CFB256_AES  # noqa: N816
usmAesBlumenthalCfb192Protocol = USM_PRIV_CFB192_AES_BLUMENTHAL  # noqa: N816
usmAesBlumenthalCfb256Protocol = USM_PRIV_CFB256_AES_BLUMENTHAL  # noqa: N816

usmKeyTypePassphrase = USM_KEY_TYPE_PASSPHRASE  # noqa: N816
usmKeyTypeMaster = USM_KEY_TYPE_MASTER  # noqa: N816
usmKeyTypeLocalized = USM_KEY_TYPE_LOCALIZED  # noqa: N816

# Old to new attribute mapping
deprecated_attributes = {
    "usmHMACMD5AuthProtocol": "USM_AUTH_HMAC96_MD5",
    "usmHMACSHAAuthProtocol": "USM_AUTH_HMAC96_SHA",
    "usmHMAC128SHA224AuthProtocol": "USM_AUTH_HMAC128_SHA224",
    "usmHMAC192SHA256AuthProtocol": "USM_AUTH_HMAC192_SHA256",
    "usmHMAC256SHA384AuthProtocol": "USM_AUTH_HMAC256_SHA384",
    "usmHMAC384SHA512AuthProtocol": "USM_AUTH_HMAC384_SHA512",
    "usmNoAuthProtocol": "USM_AUTH_NONE",
    "usmDESPrivProtocol": "USM_PRIV_CBC56_DES",
    "usm3DESEDEPrivProtocol": "USM_PRIV_CBC168_3DES",
    "usmAesCfb128Protocol": "USM_PRIV_CFB128_AES",
    "usmAesBlumenthalCfb192Protocol": "USM_PRIV_CFB192_AES_BLUMENTHAL",
    "usmAesBlumenthalCfb256Protocol": "USM_PRIV_CFB256_AES_BLUMENTHAL",
    "usmAesCfb192Protocol": "USM_PRIV_CFB192_AES",
    "usmAesCfb256Protocol": "USM_PRIV_CFB256_AES",
    "usmNoPrivProtocol": "USM_PRIV_NONE",
    "usmKeyTypePassphrase": "USM_KEY_TYPE_PASSPHRASE",
    "usmKeyTypeMaster": "USM_KEY_TYPE_MASTER",
    "usmKeyTypeLocalized": "USM_KEY_TYPE_LOCALIZED",
}


def __getattr__(attr: str):
    if new_attr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {new_attr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[new_attr]
    raise AttributeError(f"module '{__name__}' has no attribute '{attr}'")


__all__.extend(
    [
        "usm3DESEDEPrivProtocol",
        "usmAesCfb128Protocol",
        "usmAesCfb192Protocol",
        "usmAesCfb256Protocol",
        "usmAesBlumenthalCfb192Protocol",
        "usmAesBlumenthalCfb256Protocol",
        "usmDESPrivProtocol",
        "usmHMACMD5AuthProtocol",
        "usmHMACSHAAuthProtocol",
        "usmHMAC128SHA224AuthProtocol",
        "usmHMAC192SHA256AuthProtocol",
        "usmHMAC256SHA384AuthProtocol",
        "usmHMAC384SHA512AuthProtocol",
        "usmNoAuthProtocol",
        "usmNoPrivProtocol",
        "usmKeyTypePassphrase",
        "usmKeyTypeMaster",
        "usmKeyTypeLocalized",
    ]
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
"""
This module initializes the high-level API for SNMP version 3 architecture using asyncio.

Imports:
    - All entities from `pysnmp.entity.engine`
    - All entities from `pysnmp.hlapi.v3arch.asyncio` (default asyncio-based API)
    - All entities from `pysnmp.hlapi.v3arch.asyncio.auth`
    - All entities from `pysnmp.hlapi.v3arch.asyncio.context`
    - All entities from `pysnmp.proto.rfc1902`
    - Specific entities (`EndOfMibView`, `NoSuchInstance`, `NoSuchObject`) from `pysnmp.proto.rfc1905`
    - All entities from `pysnmp.smi.rfc1902`

File Information:
    - Part of the pysnmp software.
    - Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
    - License: https://www.pysnmp.com/pysnmp/license.html
"""
from pysnmp.entity.engine import *
from pysnmp.hlapi.v3arch.asyncio import *  # default is asyncio-based API
from pysnmp.hlapi.v3arch.asyncio.auth import *
from pysnmp.hlapi.v3arch.asyncio.context import *
from pysnmp.proto.rfc1902 import *
from pysnmp.proto.rfc1905 import EndOfMibView, NoSuchInstance, NoSuchObject
from pysnmp.smi.rfc1902 import *



"""
Custom ContextEngineId and ContextName
++++++++++++++++++++++++++++++++++++++

Send a SNMP GET request with the following options:

* with SNMPv3 with user 'usr-md5-none', SHA auth and no privacy protocols
* for MIB instance identified by
* contextEngineId: 0x80004fb805636c6f75644dab22cc,
* contextName: da761cfc8c94d3aceef4f60f049105ba
* over IPv4/UDP
* to an Agent at 127.0.0.1:161
* for an OID in tuple form

This script performs similar to the following Net-SNMP command:

| $ snmpget -v3 -l authNoPriv -u usr-md5-none -A authkey1 -E 80004fb805636c6f75644dab22cc -n da761cfc8c94d3aceef4f60f049105ba -ObentU 127.0.0.1:161  1.3.6.1.2.1.1.1.0

"""  #

from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import cmdgen
from pysnmp.proto import rfc1902

# Create SNMP engine instance
snmpEngine = engine.SnmpEngine()

#
# SNMPv3/USM setup
#

# user: usr-md5-none, auth: MD5, priv: NONE
config.add_v3_user(snmpEngine, "usr-md5-none", config.USM_AUTH_HMAC96_MD5, "authkey1")
config.add_target_parameters(snmpEngine, "my-creds", "usr-md5-none", "authNoPriv")

#
# Setup transport endpoint and bind it with security settings yielding
# a target name
#

# UDP/IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)
config.add_target_address(
    snmpEngine, "my-router", udp.DOMAIN_NAME, ("127.0.0.1", 161), "my-creds"
)


# Error/response receiver
# noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
def cbFun(
    snmpEngine,
    sendRequestHandle,
    errorIndication,
    errorStatus,
    errorIndex,
    varBinds,
    cbCtx,
):
    if errorIndication:
        print(errorIndication)
    elif errorStatus:
        print(
            f"{errorStatus.prettyPrint()} at {varBinds[int(errorIndex) - 1][0] if errorIndex else '?'}"
        )
    else:
        for oid, val in varBinds:
            print(f"{oid.prettyPrint()} = {val.prettyPrint()}")


# Prepare and send a request message, pass custom ContextEngineId & ContextName
cmdgen.GetCommandGenerator().send_varbinds(
    snmpEngine,
    "my-router",
    # contextEngineId
    rfc1902.OctetString(hexValue="80004fb805636c6f75644dab22cc"),
    # contextName
    rfc1902.OctetString("da761cfc8c94d3aceef4f60f049105ba"),
    [((1, 3, 6, 1, 2, 1, 1, 1, 0), None)],
    cbFun,
)

# Run I/O dispatcher which would send pending queries and process responses
snmpEngine.open_dispatcher(3)

snmpEngine.close_dispatcher()



"""
Observe SNMP engine operations
++++++++++++++++++++++++++++++

Listen and respond to SNMP GET/SET/GETNEXT/GETBULK queries with
the following options:

* SNMPv3
* with USM user 'usr-md5-des', auth: MD5, priv DES or
* allow access to SNMPv2-MIB objects (1.3.6.1.2.1)
* over IPv4/UDP, listening at 127.0.0.1:161
* registers its own execution observer to snmpEngine

The following Net-SNMP command will walk this Agent:

| $ snmpwalk -v3 -u usr-md5-des -l authPriv -A authkey1 -X privkey1 localhost .1.3.6

This script will report some details on request processing as seen
by rfc3412.receiveMessage() and rfc3412.returnResponsePdu()
abstract interfaces.

"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp

# Create SNMP engine
snmpEngine = engine.SnmpEngine()


# Execution point observer setup


# Register a callback to be invoked at specified execution point of
# SNMP Engine and passed local variables at code point's local scope
# noinspection PyUnusedLocal,PyUnusedLocal
def requestObserver(snmpEngine, execpoint, variables, cbCtx):
    print("Execution point: %s" % execpoint)
    print(
        "* transportDomain: %s"
        % ".".join([str(x) for x in variables["transportDomain"]])
    )
    print(
        "* transportAddress: {} (local {})".format(
            "@".join([str(x) for x in variables["transportAddress"]]),
            "@".join([str(x) for x in variables["transportAddress"].getLocalAddress()]),
        )
    )
    print("* securityModel: %s" % variables["securityModel"])
    print("* securityName: %s" % variables["securityName"])
    print("* securityLevel: %s" % variables["securityLevel"])
    print("* contextEngineId: %s" % variables["contextEngineId"].prettyPrint())
    print("* contextName: %s" % variables["contextName"].prettyPrint())
    print("* PDU: %s" % variables["pdu"].prettyPrint())


snmpEngine.observer.register_observer(
    requestObserver, "rfc3412.receiveMessage:request", "rfc3412.returnResponsePdu"
)

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# SNMPv3/USM setup

# user: usr-md5-des, auth: MD5, priv DES
config.add_v3_user(
    snmpEngine,
    "usr-md5-des",
    config.USM_AUTH_HMAC96_MD5,
    "authkey1",
    config.USM_PRIV_CBC56_DES,
    "privkey1",
)

# Allow full MIB access for each user at VACM
config.add_vacm_user(
    snmpEngine, 3, "usr-md5-des", "authPriv", (1, 3, 6, 1, 2, 1), (1, 3, 6, 1, 2, 1)
)

# Get default SNMP context this SNMP engine serves
snmpContext = context.SnmpContext(snmpEngine)

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()
except:
    snmpEngine.observer.unregister_observer()
    snmpEngine.close_dispatcher()
    raise



"""
Multiple SNMP engines
+++++++++++++++++++++

Send multiple SNMP GET requests to multiple peers using multiple
independent SNMP engines. Deal with peers asynchronously. SNMP options
are:

* with SNMPv1, community 'public' and
  with SNMPv2c, community 'public' and
  with SNMPv3, user 'usr-md5-des', MD5 auth and DES privacy
* over IPv4/UDP and
  over IPv6/UDP
* to an Agent at demo.pysnmp.com:161 and
  to an Agent at [::1]:161
* for instances of SNMPv2-MIB::sysDescr.0 and
  SNMPv2-MIB::sysLocation.0 MIB objects

Within this script we have a single asynchronous TransportDispatcher
and a single UDP-based transport serving two independent SNMP engines.
We use a single instance of AsyncCommandGenerator with each of
SNMP Engines to communicate GET command request to remote systems.

When we receive a [response] message from remote system we use
a custom message router to choose what of the two SNMP engines
data packet should be handed over. The selection criteria we
employ here is based on peer's UDP port number. Other selection
criteria are also possible.

"""  #
import asyncio

from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    # List of targets in the following format:
    # ( ( authData, transportTarget, varNames ), ... )
    TARGETS = (
        # 1-st target (SNMPv1 over IPv4/UDP)
        (
            CommunityData("public", mpModel=0),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
            (
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0)),
            ),
        ),
        # 2-nd target (SNMPv2c over IPv4/UDP)
        (
            CommunityData("public"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),  # TODO: 1161
            (
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0)),
            ),
        ),
        # 3-nd target (SNMPv3 over IPv4/UDP)
        (
            UsmUserData("usr-md5-des", "authkey1", "privkey1"),
            await UdpTransportTarget.create(("demo.pysnmp.com", 161)),  # TODO: 2161
            (
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysLocation", 0)),
            ),
        )
        # N-th target
        # ...
    )

    # Instantiate the single transport dispatcher object
    transportDispatcher = AsyncioDispatcher()

    # Setup a custom data routing function to select snmpEngine by transportDomain
    transportDispatcher.register_routing_callback(
        lambda td, ta, d: ta[1] % 3 and "A" or "B"
    )

    snmpEngineA = SnmpEngine()
    snmpEngineIDA = snmpEngineA.snmpEngineID
    print("snmpEngineA ID: %s" % snmpEngineIDA.prettyPrint())
    snmpEngineA.register_transport_dispatcher(transportDispatcher, "A")

    snmpEngineB = SnmpEngine()
    snmpEngineIDB = snmpEngineB.snmpEngineID
    print("snmpEngineB ID: %s" % snmpEngineIDB.prettyPrint())
    snmpEngineB.register_transport_dispatcher(transportDispatcher, "B")

    for authData, transportTarget, varBinds in TARGETS:
        snmpEngine = (
            transportTarget.get_transport_info()[1][1] % 3
            and snmpEngineA
            or snmpEngineB
        )

        (errorIndication, errorStatus, errorIndex, varBindTable) = await get_cmd(
            snmpEngine, authData, transportTarget, ContextData(), *varBinds
        )
        print(
            f"snmpEngine {snmpEngine.snmpEngineID.prettyPrint()}: {authData} via {transportTarget}"
        )
        print(
            "snmpEngineA" if snmpEngine.snmpEngineID == snmpEngineIDA else "snmpEngineB"
        )

        if errorIndication:
            print(errorIndication)
            return True

        elif errorStatus:
            print(
                "{} at {}".format(
                    errorStatus.prettyPrint(),
                    errorIndex and varBindTable[int(errorIndex) - 1][0] or "?",
                )
            )
            return True

        else:
            for varBind in varBindTable:
                print(" = ".join([x.prettyPrint() for x in varBind]))

    transportDispatcher.run_dispatcher(5)


asyncio.run(run())



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
#
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.proto import errind, error
from pysnmp.proto.secmod.rfc3414.auth import base


class NoAuth(base.AbstractAuthenticationService):
    """NoAuth authentication service.

    This service does not provide any authentication.
    """

    SERVICE_ID = (1, 3, 6, 1, 6, 3, 10, 1, 1, 1)  # usmNoAuthProtocol

    def hash_passphrase(self, authKey):
        """Hash a passphrase."""
        return

    def localize_key(self, authKey, snmpEngineID):
        """Localize a key."""
        return

    # 7.2.4.2
    def authenticate_outgoing_message(self, authKey, wholeMsg):
        """Authenticate outgoing message."""
        raise error.StatusInformation(errorIndication=errind.noAuthentication)

    def authenticate_incoming_message(self, authKey, authParameters, wholeMsg):
        """Authenticate incoming message."""
        raise error.StatusInformation(errorIndication=errind.noAuthentication)



"""
Multiple SNMP USM users
+++++++++++++++++++++++

Receive SNMP TRAP/INFORM messages with the following options:

* SNMPv3
* with USM users:

  * 'usr-md5-des', auth: MD5, priv DES, ContextEngineId: 8000000001020304
  * 'usr-md5-none', auth: MD5, priv NONE, ContextEngineId: 8000000001020304
  * 'usr-md5-aes256', auth: MD5, priv AES-256, ContextEngineId: 8000000001020304
  * 'usr-sha-aes128', auth: SHA, priv AES, ContextEngineId: 8000000001020304

* over IPv4/UDP, listening at 127.0.0.1:162
* print received data on stdout

Either of the following Net-SNMP commands will send notifications to this
receiver:

| $ snmptrap -v3 -u usr-md5-des -l authPriv -A authkey1 -X privkey1 -e 8000000001020304 127.0.0.1 123 1.3.6.1.6.3.1.1.5.1
| $ snmptrap -v3 -u usr-md5-none -l authNoPriv -A authkey1 -e 8000000001020304 127.0.0.1 123 1.3.6.1.6.3.1.1.5.1
| $ snmptrap -v3 -u usr-md5-aes256 -l authPriv -A authkey1 -x AES-256 -X privkey1 -e 8000000001020304 127.0.0.1 123 1.3.6.1.6.3.1.1.5.1
| $ snmpinform -v3 -u usr-sha-aes128 -l authPriv -a SHA -A authkey1 -x AES -X privkey1 127.0.0.1 123 1.3.6.1.6.3.1.1.5.1

Note that AES-256 is not supported by Net-SNMP. You have to use PySNMP command line tools or other SNMP software that supports AES-256.

"""  #
from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import ntfrcv
from pysnmp.proto.api import v2c

# Create SNMP engine with autogenernated engineID and pre-bound
# to socket transport dispatcher
snmpEngine = engine.SnmpEngine()

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 162))
)

# SNMPv3/USM setup

# user: usr-md5-des, auth: MD5, priv DES, securityEngineId: 8000000001020304
# this USM entry is configured for TRAP receiving purposes
config.add_v3_user(
    snmpEngine,
    "usr-md5-des",
    config.USM_AUTH_HMAC96_MD5,
    "authkey1",
    config.USM_PRIV_CBC56_DES,
    "privkey1",
    securityEngineId=v2c.OctetString(hexValue="8000000001020304"),
)

# user: usr-md5-none, auth: MD5, priv NONE, securityEngineId: 8000000001020304
# this USM entry is configured for TRAP receiving purposes
config.add_v3_user(
    snmpEngine,
    "usr-md5-none",
    config.USM_AUTH_HMAC96_MD5,
    "authkey1",
    securityEngineId=v2c.OctetString(hexValue="8000000001020304"),
)

# user: usr-md5-aes256, auth: MD5, priv AES-256, securityEngineId: 8000000001020304
# this USM entry is configured for TRAP receiving purposes
config.add_v3_user(
    snmpEngine,
    "usr-md5-aes256",
    config.USM_AUTH_HMAC96_MD5,
    "authkey1",
    config.USM_PRIV_CFB256_AES,
    "privkey1",
    securityEngineId=v2c.OctetString(hexValue="8000000001020304"),
)

# user: usr-sha-aes128, auth: SHA, priv AES, securityEngineId: 8000000001020304
# this USM entry is configured for TRAP receiving purposes
config.add_v3_user(
    snmpEngine,
    "usr-sha-aes128",
    config.USM_AUTH_HMAC96_SHA,
    "authkey1",
    config.USM_PRIV_CFB128_AES,
    "privkey1",
)


# Callback function for receiving notifications
# noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
def cbFun(snmpEngine, stateReference, contextEngineId, contextName, varBinds, cbCtx):
    print(
        'Notification from ContextEngineId "{}", ContextName "{}"'.format(
            contextEngineId.prettyPrint(), contextName.prettyPrint()
        )
    )
    for name, val in varBinds:
        print(f"{name.prettyPrint()} = {val.prettyPrint()}")


# Register SNMP Application at the SNMP engine
ntfrcv.NotificationReceiver(snmpEngine, cbFun)

snmpEngine.transport_dispatcher.job_started(1)  # this job would never finish

# Run I/O dispatcher which would receive queries and send confirmations
try:
    snmpEngine.open_dispatcher()
except:
    snmpEngine.close_dispatcher()
    raise



import pytest

from pysnmp.proto.rfc1902 import Bits


def test_prettyIn_list_valid():
    Fruit = Bits.with_named_bits(apple=0, orange=1, peach=2)

    assert Fruit().prettyIn(["apple"]) == b"\x80"
    assert Fruit().prettyIn(["orange"]) == b"\x40"
    assert Fruit().prettyIn(["peach"]) == b"\x20"


def test_prettyIn_list_multiple_valid():
    Fruit = Bits.with_named_bits(apple=0, orange=1, peach=2)

    assert Fruit().prettyIn(["peach", "apple"]) == b"\xA0"


def test_prettyIn_list_invalid():
    Fruit = Bits.with_named_bits(apple=0, orange=1, peach=2)

    with pytest.raises(Exception):
        Fruit().prettyIn(["banana"])


def test_prettyIn_multiple_octets():
    Fruit = Bits.with_named_bits(
        apple=0,
        orange=1,
        peach=2,
        banana=3,
        watermelon=4,
        strawberry=5,
        mango=6,
        lemon=7,
        cherry=8,
    )

    assert len(Fruit().prettyIn(["cherry"])) > 1


def test_prettyOut_valid():
    Fruit = Bits.with_named_bits(apple=0, orange=1, peach=2)

    assert Fruit().prettyOut(b"\x20") == "peach"
    assert Fruit().prettyOut(b"\x40") == "orange"
    assert Fruit().prettyOut(b"\x80") == "apple"


def test_prettyOut_invalid():
    Fruit = Bits.with_named_bits(apple=0, orange=1, peach=2)

    assert Fruit().prettyOut(b"\x01").startswith("Unknown")



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.error import PyAsn1Error
from pysnmp import debug
from pysnmp.error import PySnmpError


class ProtocolError(PySnmpError, PyAsn1Error):
    """Base class for all SNMP protocol errors."""

    pass


# SNMP v3 exceptions


class SnmpV3Error(ProtocolError):
    """Base class for all SNMPv3 exceptions."""

    pass


class StatusInformation(SnmpV3Error):
    """SNMPv3 status information object."""

    def __init__(self, **kwargs):
        """Create an SNMPv3 status information object."""
        SnmpV3Error.__init__(self)
        self.__errorIndication = kwargs
        debug.logger & (
            debug.FLAG_DSP | debug.FLAG_MP | debug.FLAG_SM | debug.FLAG_ACL
        ) and debug.logger("StatusInformation: %s" % kwargs)

    def __str__(self):
        """Return error indication as a string."""
        return str(self.__errorIndication)

    def __getitem__(self, key):
        """Return error indication value by key."""
        return self.__errorIndication[key]

    def __contains__(self, key):
        """Return True if key is present in the error indication."""
        return key in self.__errorIndication

    def get(self, key, defVal=None):
        """Return error indication value by key or default value."""
        return self.__errorIndication.get(key, defVal)


class CacheExpiredError(SnmpV3Error):
    """SNMPv3 cache expired exception."""

    pass


class InternalError(SnmpV3Error):
    """SNMPv3 internal error exception."""

    pass


class MessageProcessingError(SnmpV3Error):
    """SNMPv3 message processing error exception."""

    pass


class RequestTimeout(SnmpV3Error):
    """SNMPv3 request timeout exception."""

    pass



from pysnmp.carrier.asyncio.dgram import udp, udp6
from pysnmp.entity import config, engine
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.hlapi.v3arch.asyncio import SnmpEngine
from pysnmp.proto.api import v2c


import asyncio
import time


# Set the port to 1611 instead of 161, because 161 is a
# privileged port and requires root access
AGENT_PORT = 1611


async def start_agent(
    enable_ipv6: bool = False,
    enable_custom_objects: bool = False,
    enable_table_creation: bool = False,
) -> SnmpEngine:
    # Create SNMP engine
    snmpEngine = engine.SnmpEngine()

    # Set up transport endpoint
    config.add_transport(
        snmpEngine,
        udp.DOMAIN_NAME,
        udp.UdpTransport().open_server_mode(("localhost", AGENT_PORT)),
    )

    if enable_ipv6:
        config.add_transport(
            snmpEngine,
            udp6.DOMAIN_NAME,
            udp6.Udp6Transport().open_server_mode(("localhost", AGENT_PORT)),
        )

    # Set up community data
    config.add_v1_system(snmpEngine, "public", "public")
    # Add SNMP v3 user
    config.add_v3_user(
        snmpEngine, "usr-none-none", config.USM_AUTH_NONE, config.USM_PRIV_NONE
    )

    config.add_v3_user(
        snmpEngine, "usr-sha-none", config.USM_AUTH_HMAC96_SHA, "authkey1"
    )

    config.add_v3_user(
        snmpEngine,
        "usr-sha-aes",
        config.USM_AUTH_HMAC96_SHA,
        "authkey1",
        config.USM_PRIV_CFB128_AES,
        "privkey1",
    )

    config.add_v3_user(
        snmpEngine,
        "usr-sha-aes256",
        config.USM_AUTH_HMAC96_SHA,
        "authkey1",
        config.USM_PRIV_CFB256_AES,
        "privkey1",
    )

    # Allow read MIB access for this user / securityModels at VACM
    config.add_vacm_user(snmpEngine, 1, "public", "noAuthNoPriv", (1, 3, 6), (1, 3, 6))
    config.add_vacm_user(snmpEngine, 2, "public", "noAuthNoPriv", (1, 3, 6), (1, 3, 6))
    config.add_vacm_user(snmpEngine, 3, "usr-none-none", "noAuthNoPriv", (1, 3, 6))
    config.add_vacm_user(snmpEngine, 3, "usr-sha-none", "authNoPriv", (1, 3, 6))
    config.add_vacm_user(snmpEngine, 3, "usr-sha-aes", "authPriv", (1, 3, 6))
    config.add_vacm_user(snmpEngine, 3, "usr-sha-aes256", "authPriv", (1, 3, 6))

    # Configure SNMP context
    snmpContext = context.SnmpContext(snmpEngine)

    if enable_custom_objects:
        # --- create custom Managed Object Instances ---
        mibBuilder = snmpContext.get_mib_instrum().get_mib_builder()

        MibScalar, MibScalarInstance = mibBuilder.import_symbols(
            "SNMPv2-SMI", "MibScalar", "MibScalarInstance"
        )

        (
            DisplayString,
            PhysAddress,
            MacAddress,
            DateAndTime,
            TextualConvention,
        ) = mibBuilder.importSymbols(
            "SNMPv2-TC",
            "DisplayString",
            "PhysAddress",
            "MacAddress",
            "DateAndTime",
            "TextualConvention",
        )

        class SlowMibScalarInstance(MibScalarInstance):
            def getValue(self, name, **context):
                time.sleep(2)  # Add a 2-second sleep
                print("sleep ended")
                return self.getSyntax().clone(f"Test agent")

            def setValue(self, value, name, **context):
                print(f"SET operation received. New value: {value}")
                return self.getSyntax().clone(value)

        # Define a fixed date and time in the DateAndTime format
        def get_fixed_date_and_time():
            # Fixed date: 2024-11-03, 19:52:40.0, UTC offset +0:0
            return [
                2024 // 256,
                2024 % 256,  # Year (2024)
                11,  # Month (November)
                3,  # Day
                19,  # Hour (7 PM UTC)
                52,  # Minute
                40,  # Second
                0,  # Deciseconds (0 for simplicity)
                ord("+"),  # UTC offset direction ('+')
                0,  # Hours offset from UTC
                0,  # Minutes offset from UTC
            ]

        # Initialize the PhysAddress object with a sample MAC address (e.g., 00:11:22:33:44:55)
        initial_phys_address = [0x00, 0x11, 0x22, 0x33, 0x44, 0x55]

        # Initialize the MacAddress object with a sample MAC address (e.g., 00:11:22:33:44:55)
        initial_mac_address = [0x00, 0x11, 0x22, 0x33, 0x44, 0x55]

        mibBuilder.export_symbols(
            "__MY_MIB",
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 1), v2c.OctetString()),
            SlowMibScalarInstance(
                (1, 3, 6, 1, 4, 1, 60069, 9, 1), (0,), v2c.OctetString()
            ),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 3), v2c.OctetString()).setMaxAccess(
                "not-accessible"
            ),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 4), v2c.OctetString()).setMaxAccess(
                "readonly"
            ),  # PySMI <1.3.0 generates such objects
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 5), DateAndTime()).setMaxAccess(
                "read-write"
            ),
            MibScalarInstance(
                (1, 3, 6, 1, 4, 1, 60069, 9, 5),
                (0,),
                DateAndTime(get_fixed_date_and_time()),
            ),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 6), PhysAddress()).setMaxAccess(
                "read-write"
            ),
            MibScalarInstance(
                (1, 3, 6, 1, 4, 1, 60069, 9, 6), (0,), PhysAddress(initial_phys_address)
            ),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 7), v2c.Integer32()).setMaxAccess(
                "read-write"
            ),
            MibScalarInstance((1, 3, 6, 1, 4, 1, 60069, 9, 7), (0,), v2c.Integer32(50)),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 8), v2c.Unsigned32()).setMaxAccess(
                "read-write"
            ),
            MibScalarInstance((1, 3, 6, 1, 4, 1, 60069, 9, 8), (0,), v2c.Unsigned32(5)),
            MibScalar((1, 3, 6, 1, 4, 1, 60069, 9, 9), MacAddress()).setMaxAccess(
                "read-write"
            ),
            MibScalarInstance(
                (1, 3, 6, 1, 4, 1, 60069, 9, 9), (0,), MacAddress(initial_mac_address)
            ),
        )

        # --- end of Managed Object Instance initialization ----

    if enable_table_creation:
        # --- define custom SNMP Table within a newly defined EXAMPLE-MIB ---

        mibBuilder = snmpContext.get_mib_instrum().get_mib_builder()

        (
            MibTable,
            MibTableRow,
            MibTableColumn,
            MibScalarInstance,
        ) = mibBuilder.import_symbols(
            "SNMPv2-SMI",
            "MibTable",
            "MibTableRow",
            "MibTableColumn",
            "MibScalarInstance",
        )

        (RowStatus,) = mibBuilder.import_symbols("SNMPv2-TC", "RowStatus")

        mibBuilder.export_symbols(
            "__EXAMPLE-MIB",
            # table object
            exampleTable=MibTable((1, 3, 6, 6, 1)).setMaxAccess("read-create"),
            # table row object, also carries references to table indices
            exampleTableEntry=MibTableRow((1, 3, 6, 6, 1, 5))
            .setMaxAccess("read-create")
            .setIndexNames((0, "__EXAMPLE-MIB", "exampleTableColumn1")),
            # table column: string index
            exampleTableColumn1=MibTableColumn(
                (1, 3, 6, 6, 1, 5, 1), v2c.OctetString()
            ).setMaxAccess("read-create"),
            # table column: string value
            exampleTableColumn2=MibTableColumn(
                (1, 3, 6, 6, 1, 5, 2), v2c.OctetString()
            ).setMaxAccess("read-create"),
            # table column: integer value with default
            exampleTableColumn3=MibTableColumn(
                (1, 3, 6, 6, 1, 5, 3), v2c.Integer32(123)
            ).setMaxAccess("read-create"),
            # table column: row status
            exampleTableStatus=MibTableColumn(
                (1, 3, 6, 6, 1, 5, 4), RowStatus("notExists")
            ).setMaxAccess("read-create"),
        )

        # --- end of custom SNMP table definition, empty table now exists ---

        # --- populate custom SNMP table with one row ---

        (
            exampleTableEntry,
            exampleTableColumn2,
            exampleTableColumn3,
            exampleTableStatus,
        ) = mibBuilder.import_symbols(
            "__EXAMPLE-MIB",
            "exampleTableEntry",
            "exampleTableColumn2",
            "exampleTableColumn3",
            "exampleTableStatus",
        )
        rowInstanceId = exampleTableEntry.getInstIdFromIndices("example record one")
        mibInstrumentation = snmpContext.get_mib_instrum()
        mibInstrumentation.write_variables(
            (exampleTableColumn2.name + rowInstanceId, "my string value"),
            (exampleTableColumn3.name + rowInstanceId, 123456),
            (exampleTableStatus.name + rowInstanceId, "createAndGo"),
        )

    # Register SNMP Applications at the SNMP engine for particular SNMP context
    cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
    cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
    cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)
    cmdrsp.SetCommandResponder(snmpEngine, snmpContext)

    # Start the event loop
    snmpEngine.transport_dispatcher.job_started(1)

    snmpEngine.open_dispatcher()

    # Wait for the agent to start
    await asyncio.sleep(1)

    # return the engine
    return snmpEngine


class AgentContextManager:
    """
    A context manager for managing the lifecycle of an SNMP test agent.

    Usage:
    async with AgentContextManager() as agent:
        # Perform operations with the agent

    When the context is entered, the agent is started using the `start_agent()` function.
    When the context is exited, the agent's transport dispatcher is stopped and closed.

    Note: The `start_agent()` function and the `transportDispatcher` attribute are not defined in this code snippet.
    """

    def __init__(
        self,
        enable_ipv6: bool = False,
        enable_custom_objects: bool = False,
        enable_table_creation: bool = False,
    ):
        self.enable_ipv6 = enable_ipv6
        self.enable_custom_objects = enable_custom_objects
        self.enable_table_creation = enable_table_creation

    async def __aenter__(self):
        self.agent = await start_agent(
            self.enable_ipv6, self.enable_custom_objects, self.enable_table_creation
        )
        return self.agent

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.agent.transport_dispatcher.job_finished(1)
        self.agent.close_dispatcher()



"""
Multiple MIB trees under distinct context names
+++++++++++++++++++++++++++++++++++++++++++++++

Listen and respond to SNMP GET/SET/GETNEXT/GETBULK queries with
the following options:

* SNMPv3
* with USM username usr-md5-none
* using two independent (empty) sets of Managed Objects addressed by
  SNMP context name `context-a` and `context-b`
* allow access to SNMPv2-MIB objects (1.3.6.1.2.1)
* over IPv4/UDP, listening at 127.0.0.1:161

Either of the following Net-SNMP commands will walk this Agent:

| $ snmpwalk -v3 -u usr-md5-none -l authNoPriv -A authkey1 -n context-a 127.0.0.1 .1.3.6
| $ snmpwalk -v3 -u usr-md5-none -l authNoPriv -A authkey1 -n context-b 127.0.0.1 .1.3.6

"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.smi import instrum, builder
from pysnmp.proto.api import v2c

# Create SNMP engine
snmpEngine = engine.SnmpEngine()

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.domainName, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# SNMPv3/USM setup

# user: usr-md5-none, auth: MD5, priv NONE
config.add_v3_user(snmpEngine, "usr-md5-none", config.USM_AUTH_HMAC96_MD5, "authkey1")

# Allow full MIB access for each user at VACM
config.add_vacm_user(
    snmpEngine, 3, "usr-md5-none", "authNoPriv", (1, 3, 6, 1, 2, 1), (1, 3, 6, 1, 2, 1)
)

# Create an SNMP context with default ContextEngineId (same as SNMP engine ID)
snmpContext = context.SnmpContext(snmpEngine)

# Create multiple independent trees of MIB managed objects (empty so far)
mibTreeA = instrum.MibInstrumController(builder.MibBuilder())
mibTreeB = instrum.MibInstrumController(builder.MibBuilder())

# Register MIB trees at distinct SNMP Context names
snmpContext.register_context_name(v2c.OctetString("context-a"), mibTreeA)
snmpContext.register_context_name(v2c.OctetString("context-b"), mibTreeB)

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()

except:
    snmpEngine.close_dispatcher()
    raise



"""
This module is necessary to make the containing directory a package.

By including this file, Python will treat the directory as a package,
allowing modules within the directory to be imported.
"""



"""
Implementing scalar MIB objects
+++++++++++++++++++++++++++++++

Listen and respond to SNMP GET/SET/GETNEXT/GETBULK queries with
the following options:

* SNMPv2c
* with SNMP community "public"
* serving custom Managed Object Instance defined within this script
* allow read access only to the subtree where the custom MIB object resides
* over IPv4/UDP, listening at 127.0.0.1:161

The following Net-SNMP commands will walk this Agent:

| $ snmpwalk -v2c -c public 127.0.0.1 .1.3.6

"""  #
import sys
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.proto.api import v2c

# Create SNMP engine
snmpEngine = engine.SnmpEngine()

# Transport setup

# UDP over IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpTransport().open_server_mode(("127.0.0.1", 161))
)

# SNMPv2c setup

# SecurityName <-> CommunityName mapping.
config.add_v1_system(snmpEngine, "my-area", "public")

# Allow read MIB access for this user / securityModels at VACM
config.add_vacm_user(snmpEngine, 2, "my-area", "noAuthNoPriv", (1, 3, 6, 5))

# Create an SNMP context
snmpContext = context.SnmpContext(snmpEngine)

# --- create custom Managed Object Instance ---

mibBuilder = snmpContext.get_mib_instrum().get_mib_builder()

MibScalar, MibScalarInstance = mibBuilder.import_symbols(
    "SNMPv2-SMI", "MibScalar", "MibScalarInstance"
)


class MyStaticMibScalarInstance(MibScalarInstance):
    # noinspection PyUnusedLocal,PyUnusedLocal
    def getValue(self, name, **context):
        return self.getSyntax().clone(
            f"Python {sys.version} running on a {sys.platform} platform"
        )


mibBuilder.export_symbols(
    "__MY_MIB",
    MibScalar((1, 3, 6, 5, 1), v2c.OctetString()),
    MyStaticMibScalarInstance((1, 3, 6, 5, 1), (0,), v2c.OctetString()),
)

# --- end of Managed Object Instance initialization ----

# Register SNMP Applications at the SNMP engine for particular SNMP context
cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
snmpEngine.transport_dispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    snmpEngine.open_dispatcher()
except:
    snmpEngine.close_dispatcher()
    raise



"""
Bulk walk Agent MIB (SNMPv2c)
+++++++++++++++++++++++++++++

Perform SNMP GETBULK operation with the following options:

* with SNMPv2c, community 'public'
* over IPv4/UDP
* to an Agent at demo.pysnmp.com:161
* for OID in tuple form
* with non-repeaters=0 and max-repeaters=25

This script performs similar to the following Net-SNMP command:

| $ snmpbulkwalk -v2c -c public -ObentU -Cn0 -Cr25 demo.pysnmp.com 1.3.6

"""  #
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.carrier.asyncio.dgram import udp
from pyasn1.codec.ber import encoder, decoder
from pysnmp.proto.api import v2c

# SNMP table header
headVars = [v2c.ObjectIdentifier((1, 3, 6))]

# Build PDU
reqPDU = v2c.GetBulkRequestPDU()
v2c.apiBulkPDU.set_defaults(reqPDU)
v2c.apiBulkPDU.set_non_repeaters(reqPDU, 0)
v2c.apiBulkPDU.set_max_repetitions(reqPDU, 25)
v2c.apiBulkPDU.set_varbinds(reqPDU, [(x, v2c.null) for x in headVars])

# Build message
reqMsg = v2c.Message()
v2c.apiMessage.set_defaults(reqMsg)
v2c.apiMessage.set_community(reqMsg, "public")
v2c.apiMessage.set_pdu(reqMsg, reqPDU)


# noinspection PyUnusedLocal
def cbRecvFun(
    transportDispatcher,
    transportDomain,
    transportAddress,
    wholeMsg,
    reqPDU=reqPDU,
    headVars=headVars,
):
    while wholeMsg:
        rspMsg, wholeMsg = decoder.decode(wholeMsg, asn1Spec=v2c.Message())

        rspPDU = v2c.apiMessage.get_pdu(rspMsg)

        # Match response to request
        if v2c.apiBulkPDU.get_request_id(reqPDU) == v2c.apiBulkPDU.get_request_id(
            rspPDU
        ):
            # Format var-binds table
            varBindTable = v2c.apiBulkPDU.get_varbind_table(reqPDU, rspPDU)

            # Check for SNMP errors reported
            errorStatus = v2c.apiBulkPDU.get_error_status(rspPDU)
            if errorStatus and errorStatus != 2:
                errorIndex = v2c.apiBulkPDU.get_error_index(rspPDU)
                print(
                    "{} at {}".format(
                        errorStatus.prettyPrint(),
                        errorIndex and varBindTable[int(errorIndex) - 1] or "?",
                    )
                )
                transportDispatcher.job_finished(1)
                break

            # Report SNMP table
            for tableRow in varBindTable:
                for name, val in tableRow:
                    print(
                        "from: {}, {} = {}".format(
                            transportAddress, name.prettyPrint(), val.prettyPrint()
                        )
                    )

            # Stop on EOM
            for oid, val in varBindTable[-1]:
                if not isinstance(val, v2c.Null):
                    break

            else:
                transportDispatcher.job_finished(1)
                continue

            # Generate request for next row
            v2c.apiBulkPDU.set_varbinds(
                reqPDU, [(x, v2c.null) for x, y in varBindTable[-1]]
            )

            v2c.apiBulkPDU.set_request_id(reqPDU, v2c.getNextRequestID())

            transportDispatcher.send_message(
                encoder.encode(reqMsg), transportDomain, transportAddress
            )

    return wholeMsg


transportDispatcher = AsyncioDispatcher()

transportDispatcher.register_recv_callback(cbRecvFun)

transportDispatcher.register_transport(
    udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)

transportDispatcher.send_message(
    encoder.encode(reqMsg), udp.DOMAIN_NAME, ("demo.pysnmp.com", 161)
)

transportDispatcher.job_started(1)

# Dispatcher will finish as job#1 counter reaches zero
transportDispatcher.run_dispatcher(3)

transportDispatcher.close_dispatcher()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
import warnings

from pyasn1.type import constraint, namedtype, namedval, tag, univ
from pysnmp.proto import error, rfc1155

__all__ = [
    "Opaque",
    "TimeTicks",
    "Bits",
    "Integer",
    "OctetString",
    "IpAddress",
    "Counter64",
    "Unsigned32",
    "Gauge32",
    "Integer32",
    "ObjectIdentifier",
    "Counter32",
    "Null",
]


class Null(univ.Null):
    """Creates an instance of SNMP Null class.

    :py:class:`~pysnmp.proto.rfc1902.Null` type represents the absence
    of value.

    Parameters
    ----------
    initializer: str
        Python string object. Must be an empty string.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Null('')
        Null('')
        >>>
    """


class Integer32(univ.Integer):
    """Creates an instance of SNMP Integer32 class.

    :py:class:`~pysnmp.proto.rfc1902.Integer32` type represents
    integer-valued information between -2147483648 to 2147483647
    inclusive (:RFC:`1902#section-7.1.1`). This type is indistinguishable
    from the :py:class:`~pysnmp.proto.rfc1902.Integer` type.
    The :py:class:`~pysnmp.proto.rfc1902.Integer32` type may be sub-typed
    to be more constrained than the base
    :py:class:`~pysnmp.proto.rfc1902.Integer32` type.

    Parameters
    ----------
    initializer : int
        Python integer in range between -2147483648 to 2147483647 inclusive
        or :py:class:`~pysnmp.proto.rfc1902.Integer32`.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Integer32(1234)
        Integer32(1234)
        >>> Integer32(1) > 2
        True
        >>> Integer32(1) + 1
        Integer32(2)
        >>> int(Integer32(321))
        321
        >>> SmallInteger = Integer32.with_range(1,3)
        >>> SmallInteger(1)
        Integer32(1)
        >>> DiscreetInteger = Integer32.with_values(4, 8, 1)
        >>> DiscreetInteger(4)
        Integer32(4)
        >>>

    """

    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        -2147483648, 2147483647
    )  # noqa: N815

    @classmethod
    def with_values(cls, *values):
        """Creates a subclass with discreet values constraint."""

        class X(cls):
            subtypeSpec = cls.subtypeSpec + constraint.SingleValueConstraint(
                *values
            )  # noqa: N815

        X.__name__ = cls.__name__
        return X

    @classmethod
    def with_range(cls, minimum, maximum):
        """Creates a subclass with value range constraint."""

        class X(cls):
            subtypeSpec = cls.subtypeSpec + constraint.ValueRangeConstraint(
                minimum, maximum
            )  # noqa: N815

        X.__name__ = cls.__name__
        return X


class Integer(Integer32):
    """Creates an instance of SNMP INTEGER class.

    The :py:class:`~pysnmp.proto.rfc1902.Integer` type represents
    integer-valued information as named-number enumerations
    (:RFC:`1902#section-7.1.1`). This type inherits and is indistinguishable
    from :py:class:`~pysnmp.proto.rfc1902.Integer32` class.
    The :py:class:`~pysnmp.proto.rfc1902.Integer` type may be sub-typed
    to be more constrained than the base
    :py:class:`~pysnmp.proto.rfc1902.Integer` type.

    Parameters
    ----------
    initializer : int
        Python integer in range between -2147483648 to 2147483647 inclusive
        or :py:class:`~pysnmp.proto.rfc1902.Integer`  class instance.
        In case of named-numbered enumerations, initialization is also
        possible by enumerated literal.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Integer(1234)
        Integer(1234)
        >>> Integer(1) > 2
        True
        >>> Integer(1) + 1
        Integer(2)
        >>> int(Integer(321))
        321
        >>> SomeState = Integer.with_named_values(enable=1, disable=0)
        >>> SomeState(1)
        Integer('enable')
        >>> int(SomeState('disable'))
        0
        >>>

    """

    @classmethod
    def with_named_values(cls, **values):
        """Create a subclass with discreet named values constraint.

        Reduce fully duplicate enumerations along the way.
        """
        enums = set(cls.namedValues.items())
        enums.update(values.items())

        class X(cls):
            namedValues = namedval.NamedValues(*enums)
            subtypeSpec = cls.subtypeSpec + constraint.SingleValueConstraint(
                *values.values()
            )  # noqa: N815

        X.__name__ = cls.__name__
        return X


class OctetString(univ.OctetString):
    r"""Creates an instance of SNMP OCTET STRING class.

    The :py:class:`~pysnmp.proto.rfc1902.OctetString` type represents
    arbitrary binary or text data (:RFC:`1902#section-7.1.2`).
    It may be sub-typed to be constrained in size.

    Parameters
    ----------
    strValue : str
        Python string or :py:class:`~pysnmp.proto.rfc1902.OctetString`
        class instance.

    Other parameters
    ----------------
    hexValue : str
        Python string representing octets in a hexadecimal notation
        (e.g. DEADBEEF).

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> OctetString('some apples')
        OctetString('some apples')
        >>> OctetString('some apples') + ' and oranges'
        OctetString('some apples and oranges')
        >>> str(OctetString('some apples'))
        'some apples'
        >>> SomeString = OctetString.with_size(3, 12)
        >>> str(SomeString(hexValue='deadbeef'))
        '\xde\xad\xbe\xef'
        >>>

    """

    subtypeSpec = univ.OctetString.subtypeSpec + constraint.ValueSizeConstraint(
        0, 65535
    )  # noqa: N815

    # rfc1902 uses a notion of "fixed length string" what might mean
    # having zero-range size constraint applied. The following is
    # supposed to be used for setting and querying this property.

    fixed_length = None

    def set_fixed_length(self, value):
        """Set fixed length."""
        self.fixed_length = value
        return self

    def is_fixed_length(self):
        """Return if fixed length."""
        return self.fixed_length is not None

    def get_fixed_length(self):
        """Return fixed length."""
        return self.fixed_length

    def clone(self, *args, **kwargs):
        """Clone the data."""
        return univ.OctetString.clone(self, *args, **kwargs).set_fixed_length(
            self.get_fixed_length()
        )

    def subtype(self, *args, **kwargs):
        """Subtype the data."""
        return univ.OctetString.subtype(self, *args, **kwargs).set_fixed_length(
            self.get_fixed_length()
        )

    @classmethod
    def with_size(cls, minimum, maximum):
        """Creates a subclass with value size constraint."""

        class X(cls):
            subtypeSpec = cls.subtypeSpec + constraint.ValueSizeConstraint(
                minimum, maximum
            )  # noqa: N815

        X.__name__ = cls.__name__
        return X

    # Compatibility API
    deprecated_attributes = {
        "setFixedLength": "set_fixed_length",
        "isFixedLength": "is_fixed_length",
        "getFixedLength": "get_fixed_length",
        "fixedLength": "fixed_length",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{attr}'"
        )


class ObjectIdentifier(univ.ObjectIdentifier):
    """Creates an instance of SNMP OBJECT IDENTIFIER class.

    The :py:class:`~pysnmp.proto.rfc1902.ObjectIdentifier` type represents
    administratively assigned names (:RFC:`1902#section-7.1.3`).
    Supports sequence protocol where elements are integer sub-identifiers.

    Parameters
    ----------
    initializer: tuple, str
        Python tuple of up to 128 integers in range between 0 to 4294967295
        inclusive or Python string containing OID in "dotted" form or
        :py:class:`~pysnmp.proto.rfc1902.ObjectIdentifier`.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> ObjectIdentifier((1, 3, 6))
        ObjectIdentifier('1.3.6')
        >>> ObjectIdentifier('1.3.6')
        ObjectIdentifier('1.3.6')
        >>> tuple(ObjectIdentifier('1.3.6'))
        (1, 3, 6)
        >>> str(ObjectIdentifier('1.3.6'))
        '1.3.6'
        >>>

    """


class IpAddress(OctetString):
    r"""Creates an instance of SNMP IpAddress class.

    The :py:class:`~pysnmp.proto.rfc1902.IpAddress` class represents
    a 32-bit internet address as an OCTET STRING of length 4, in network
    byte-order (:RFC:`1902#section-7.1.5`).

    Parameters
    ----------
    strValue : str
        The same as :py:class:`~pysnmp.proto.rfc1902.OctetString`,
        additionally IPv4 address in dotted notation ('127.0.0.1').

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> IpAddress('127.0.0.1')
        IpAddress(hexValue='7f000001')
        >>> str(IpAddress(hexValue='7f000001'))
        '\x7f\x00\x00\x01'
        >>> IpAddress('\x7f\x00\x00\x01')
        IpAddress(hexValue='7f000001')
        >>>

    """

    tagSet = OctetString.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x00)
    )
    subtypeSpec = OctetString.subtypeSpec + constraint.ValueSizeConstraint(
        4, 4
    )  # noqa: N815
    fixed_length = 4

    def prettyIn(self, value):  # noqa: N802
        """Convert string to IP address."""
        if isinstance(value, str) and len(value) != 4:
            try:
                value = [int(x) for x in value.split(".")]
            except:
                raise error.ProtocolError("Bad IP address syntax %s" % value)
        value = OctetString.prettyIn(self, value)
        if len(value) != 4:
            raise error.ProtocolError("Bad IP address syntax")
        return value

    def prettyOut(self, value):  # noqa: N802
        """Convert IP address to a string."""
        if value:
            return ".".join(["%d" % x for x in self.__class__(value).asNumbers()])
        else:
            return ""


class Counter32(univ.Integer):
    """Creates an instance of SNMP Counter32 class.

    :py:class:`~pysnmp.proto.rfc1902.Counter32` type represents
    a non-negative integer which monotonically increases until it
    reaches a maximum value of 4294967295, when it wraps around and
    starts increasing again from zero (:RFC:`1902#section-7.1.6`).

    Parameters
    ----------
    initializer : int
        Python integer in range between 0 to 4294967295 inclusive
        or any :py:class:`~pysnmp.proto.rfc1902.Integer`-based class.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Counter32(1234)
        Counter32(1234)
        >>> Counter32(1) + 1
        Counter32(2)
        >>> int(Counter32(321))
        321
        >>>

    """

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x01)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class Gauge32(univ.Integer):
    """Creates an instance of SNMP Gauge32 class.

    :py:class:`~pysnmp.proto.rfc1902.Gauge32` type represents
    a non-negative integer, which may increase or decrease, but shall
    never exceed a maximum value. The maximum value can not be greater
    than 4294967295 (:RFC:`1902#section-7.1.7`).

    Parameters
    ----------
    initializer : int
        Python integer in range between 0 to 4294967295 inclusive
        or any :py:class:`~pysnmp.proto.rfc1902.Integer`-based class.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Gauge32(1234)
        Gauge32(1234)
        >>> Gauge32(1) + 1
        Gauge32(2)
        >>> int(Gauge32(321))
        321
        >>>

    """

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x02)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class Unsigned32(univ.Integer):
    """Creates an instance of SNMP Unsigned32 class.

    :py:class:`~pysnmp.proto.rfc1902.Unsigned32` type represents
    integer-valued information between 0 and 4294967295
    (:RFC:`1902#section-7.1.11`).

    Parameters
    ----------
    initializer : int
        Python integer in range between 0 to 4294967295 inclusive
        or any :py:class:`~pysnmp.proto.rfc1902.Integer`-based class.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Unsigned32(1234)
        Unsigned32(1234)
        >>> Unsigned32(1) + 1
        Unsigned32(2)
        >>> int(Unsigned32(321))
        321
        >>>

    """

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x02)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class TimeTicks(univ.Integer):
    """Creates an instance of SNMP TimeTicks class.

    :py:class:`~pysnmp.proto.rfc1902.TimeTicks` type represents
    a non-negative integer which represents the time, modulo 4294967296,
    in hundredths of a second between two epochs (:RFC:`1902#section-7.1.8`).

    Parameters
    ----------
    initializer : int
        Python integer in range between 0 to 4294967295 inclusive
        or any :py:class:`~pysnmp.proto.rfc1902.Integer`-based class.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> TimeTicks(1234)
        TimeTicks(1234)
        >>> TimeTicks(1) + 1
        TimeTicks(2)
        >>> int(TimeTicks(321))
        321
        >>>

    """

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x03)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 4294967295
    )  # noqa: N815


class Opaque(univ.OctetString):
    r"""Creates an instance of SNMP Opaque class.

    The :py:class:`~pysnmp.proto.rfc1902.Opaque` type supports the
    capability to pass arbitrary ASN.1 syntax.  A value is encoded
    using the ASN.1 BER into a string of octets.  This, in turn, is
    encoded as an OCTET STRING, in effect "double-wrapping" the original
    ASN.1 value (:RFC:`1902#section-7.1.9`).

    Parameters
    ----------
    strValue : str
        Python string or :py:class:`~pysnmp.proto.rfc1902.OctetString`-based
        class instance.

    Other parameters
    ----------------
    hexValue : str
        Python string representing octets in a hexadecimal notation
        (e.g. DEADBEEF).

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Opaque('some apples')
        Opaque('some apples')
        >>> Opaque('some apples') + ' and oranges'
        Opaque('some apples and oranges')
        >>> str(Opaque('some apples'))
        'some apples'
        >>> str(Opaque(hexValue='deadbeef'))
        '\xde\xad\xbe\xef'
        >>>

    """

    tagSet = univ.OctetString.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x04)
    )


class Counter64(univ.Integer):
    """Creates an instance of SNMP Counter64 class.

    :py:class:`~pysnmp.proto.rfc1902.Counter64` type represents
    a non-negative integer which monotonically increases until it reaches
    a maximum value of 18446744073709551615, when it wraps around and starts
    increasing again from zero (:RFC:`1902#section-7.1.10`).

    Parameters
    ----------
    initializer : int
        Python integer in range between 0 to 4294967295 inclusive
        or any :py:class:`~pysnmp.proto.rfc1902.Integer`-based class.

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> Counter64(1234)
        Counter64(1234)
        >>> Counter64(1) + 1
        Counter64(2)
        >>> int(Counter64(321))
        321
        >>>

    """

    tagSet = univ.Integer.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassApplication, tag.tagFormatSimple, 0x06)
    )
    subtypeSpec = univ.Integer.subtypeSpec + constraint.ValueRangeConstraint(
        0, 18446744073709551615
    )  # noqa: N815


class Bits(OctetString):
    r"""Creates an instance of SNMP BITS class.

    The :py:class:`~pysnmp.proto.rfc1902.Bits` type represents
    an enumeration of named bits. This collection is assigned non-negative,
    contiguous values, starting at zero. Only those named-bits so enumerated
    may be present in a value (:RFC:`1902#section-7.1.4`).

    The bits are named and identified by their position in the octet string.
    Position zero is the high order (or left-most) bit in the first octet of
    the string. Position 7 is the low order (or right-most) bit of the first
    octet of the string. Position 8 is the high order bit in the second octet
    of the string, and so on
    (`BITS Pseudotype <https://tools.ietf.org/html/draft-perkins-bits-00>`_).

    Parameters
    ----------
    strValue : str, tuple
        Sequence of bit names or a Python string (as a raw data) or
        :py:class:`~pysnmp.proto.rfc1902.OctetString` class instance.

    Other parameters
    ----------------
    hexValue : str
        Python string representing octets in a hexadecimal notation
        (e.g. DEADBEEF).

    Raises
    ------
        PyAsn1Error :
            On constraint violation or bad initializer.

    Examples
    --------
        >>> from pysnmp.proto.rfc1902 import *
        >>> SomeBits = Bits.with_named_bits(apple=0, orange=1, peach=2)
        >>> SomeBits(('apple', 'orange')).prettyPrint()
        'apple, orange'
        >>> SomeBits(('apple', 'orange'))
        Bits(hexValue='c0')
        >>> SomeBits('\x80')
        Bits(hexValue='80')
        >>> SomeBits(hexValue='80')
        Bits(hexValue='80')
        >>> SomeBits(hexValue='80').prettyPrint()
        'apple'
        >>>

    """

    namedValues: namedval.NamedValues = namedval.NamedValues()

    def __new__(cls, *args, **kwargs):
        """Create a new instance of the class."""
        if "namedValues" in kwargs:
            Bits = cls.with_named_bits(**dict(kwargs.pop("namedValues")))
            return Bits(*args, **kwargs)

        return OctetString.__new__(cls)

    def prettyIn(self, bits):  # noqa: N802
        """Return raw bitstring."""
        if not isinstance(bits, (tuple, list)):
            return OctetString.prettyIn(self, bits)  # raw bitstring
        octets = []
        for bit in bits:  # tuple of named bits
            v = self.namedValues.getValue(bit)
            if v is None:
                raise error.ProtocolError("Unknown named bit %s" % bit)
            d, m = divmod(v, 8)
            if d >= len(octets):
                octets.extend([0] * (d - len(octets) + 1))
            octets[d] |= 0x01 << (7 - m)
        return OctetString.prettyIn(self, octets)

    def prettyOut(self, value):  # noqa: N802
        """Return named bits."""
        names = []
        ints = self.__class__(value).asNumbers()
        for i, v in enumerate(ints):
            v = ints[i]
            j = 7
            while j >= 0:
                if v & (0x01 << j):
                    name = self.namedValues.getName(i * 8 + 7 - j)
                    if name is None:
                        name = f"UnknownBit-{i * 8 + 7 - j}"
                    names.append(name)
                j -= 1
        return ", ".join([str(x) for x in names])

    @classmethod
    def with_named_bits(cls, **values):
        """Creates a subclass with discreet named bits constraint.

        Reduce fully duplicate enumerations along the way.
        """
        enums = set(cls.namedValues.items())
        enums.update(values.items())

        class X(cls):
            namedValues = namedval.NamedValues(*enums)

        X.__name__ = cls.__name__
        return X


class ObjectName(univ.ObjectIdentifier):
    pass


class SimpleSyntax(rfc1155.TypeCoercionHackMixIn, univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("integer-value", Integer()),
        namedtype.NamedType("string-value", OctetString()),
        namedtype.NamedType("objectID-value", univ.ObjectIdentifier()),
    )


class ApplicationSyntax(rfc1155.TypeCoercionHackMixIn, univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("ipAddress-value", IpAddress()),
        namedtype.NamedType("counter-value", Counter32()),
        namedtype.NamedType("timeticks-value", TimeTicks()),
        namedtype.NamedType("arbitrary-value", Opaque()),
        namedtype.NamedType("big-counter-value", Counter64()),
        # This conflicts with Counter32
        # namedtype.NamedType('unsigned-integer-value', Unsigned32()),
        namedtype.NamedType("gauge32-value", Gauge32()),
    )  # BITS misplaced?


class ObjectSyntax(univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("simple", SimpleSyntax()),
        namedtype.NamedType("application-wide", ApplicationSyntax()),
    )
